import matplotlib
matplotlib.use('agg')
import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import GroupShuffleSplit
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.calibration import CalibratedClassifierCV
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import cross_val_score,cross_val_predict
from sklearn.model_selection import GroupKFold
from scipy.stats import pearsonr
import seaborn as sns
from microbiome.helpers import *
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D 
import math
import pathlib
from sklearn.model_selection import GroupKFold
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import r2_score
from sklearn.model_selection import GroupKFold
from sklearn.model_selection import GroupShuffleSplit
from catboost import CatBoostRegressor
import pandas as pd
import scipy as sp
import scipy.stats as stats
import seaborn as sns

from sklearn.neighbors import KNeighborsClassifier

def zscore_analysis(df_all, z_col_name, hue_col, cross_limit=2, plot=True):
    df = df_all.copy()
    df["month"] = df.age_at_collection//30
    
    outliers_above, outliers_below = [], []
    
    df_mean_std = df[["month", z_col_name]].groupby(["month"]).agg([np.mean, np.std]).reset_index()

    def fun(row):
        mean = df_mean_std[df_mean_std.month==row["month"]][[(z_col_name, 'mean')]].values[0][0]
        std  = df_mean_std[df_mean_std.month==row["month"]][[(z_col_name, 'std')]].values[0][0]
        row[f"z_{z_col_name}"] = (row[z_col_name] - mean)/std
        return row

    df = df.apply(lambda row: fun(row), axis=1)
    df = df.sort_values(by=["month"])
    
    # plot for individual subjects
    fig, ax = plt.subplots(figsize=(20,10))

    x_lim = 39
    y_lim = 6

    ax.plot(np.linspace(0, x_lim, num=x_lim), np.zeros(x_lim), c="k", lw=2)
    for i in range(1, y_lim):
        ax.plot(np.linspace(0, x_lim, num=x_lim), np.ones(x_lim)*i, c="gray", lw=2, linestyle="--")
        ax.plot(np.linspace(0, x_lim, num=x_lim), -np.ones(x_lim)*i, c="gray", lw=2, linestyle="--")

    for s in df.subjectID.unique():
        df_subj = df[df.subjectID==s]
        #df_subj = df_subj[(df_subj.z_weight_growth_pace_during_three_years.max()<=cross_limit)&(df_subj.z_weight_growth_pace_during_three_years.min()>=-cross_limit)]
        
        if -cross_limit <= df_subj[f"z_{z_col_name}"].max()<=cross_limit and \
            -cross_limit <= df_subj[f"z_{z_col_name}"].min()<=cross_limit: 
            ax.plot(df_subj.month, df_subj[f"z_{z_col_name}"], marker="o", label=s)
        else:
            if df_subj[f"z_{z_col_name}"].max() < -cross_limit or \
                df_subj[f"z_{z_col_name}"].min() < -cross_limit:
                outliers_below.append(s)
            elif df_subj[f"z_{z_col_name}"].max() > cross_limit or \
                df_subj[f"z_{z_col_name}"].min() > cross_limit:
                outliers_above.append(s)
            else:
                pass
                #print("Check this case?")
    
    ax.set_ylim((-y_lim, y_lim));ax.set_xlim((0, x_lim))
    ax.legend(ncol=12, bbox_to_anchor=(1., -.05))
    
    # plot mean+std of z-score
    fig, ax = plt.subplots(figsize=(20,10))

    x_lim = 39
    y_lim = 3

    ax.plot(np.linspace(0, x_lim, num=x_lim), np.zeros(x_lim), c="k", lw=2)
    for i in range(1, y_lim):
        ax.plot(np.linspace(0, x_lim, num=x_lim), np.ones(x_lim)*i, c="gray", lw=2, linestyle="--")
        ax.plot(np.linspace(0, x_lim, num=x_lim), -np.ones(x_lim)*i, c="gray", lw=2, linestyle="--")
    ax = sns.pointplot(x="month", y=f"z_{z_col_name}", hue=hue_col, data=df, ax=ax, capsize=.2)
    #ax.scatter(df_subj.age_at_collection, df_subj.waz_last)
    ax.set_ylim((-y_lim, y_lim));ax.set_xlim((0, x_lim));

    if plot:
        plt.show()
    
    return outliers_below, outliers_above


n_splits = 5
test_size = 0.5

Regressor = CatBoostRegressor
parameters = {"loss_function": "MAE",
              "random_state": 42,
              "allow_writing_files": True,
             "verbose":False}
param_grid = { 'learning_rate': [0.5, 0.1],
               #"depth": [5, 10],
               #"iterations": [100, 500, 1000]
             }

def find_best_reference_with_least_crossings(df, feature_columns, nice_name=lambda x: x, file_directory=None):
    results = pd.DataFrame({
        "bacteria_name":[], 
        "total_num_of_crossings":[],
        "total_num_of_crossings_smooth":[], 
        "MAE":[],
        "R2":[]})
    for b_ref in feature_columns:
        total_num_of_crossings1, total_num_of_crossings2, mae, r2 = plot_shap_abundances_and_ratio(df, feature_columns, b_ref, nice_name, Regressor=Regressor, parameters=parameters, n_splits=n_splits, 
                                                                                                    file_name=f"{file_directory}/{b_ref.replace(';','_')}.png", plot=False);
        results = results.append({
            "bacteria_name":b_ref, 
            "total_num_of_crossings":total_num_of_crossings1,
            "total_num_of_crossings_smooth":total_num_of_crossings2,
            "MAE":mae,
            "r2":r2
        }, ignore_index=True)
        print(b_ref, "-->", total_num_of_crossings1, total_num_of_crossings2)
    
    file_name = f"{file_directory}/crossings.xls" or "crossings.xls"
    results.to_csv(file_name, sep="\t", index=False)
    return results
    

def plot_shap_abundances_and_ratio(df, important_features, bacteria_name, nice_name, file_name, Regressor=Regressor, parameters=parameters, n_splits=n_splits, plot=False, patent=False, time_unit_size=30, time_unit_name="months"):
    """ Plot the shap values and abundances and ratios

    Plot that is used when working with ratios. We wanted to see plots side-by-side that represent the abundance of bacteria. On the other plot
    the ratio of these two bacterias. We wanted to understand better this mapping of abundances to ratios. At the end we concluded that we want 
    to use the bacteria in the numerator that has the least number of crossings when we look at the 2 abundances.
    
    color1 = other abundance
    color2 = abundance reference
    color3 = crossing
    color4 = zero line
    colorbg = bg color
    """
    if not patent:
        color1 = "green"
        color2 = "blue"
        color3 = "red"
        color4 = "orange"
        color5 = "gray"
        colorbg = "white"
    else:
        color1 = "black"
        color2 = "white"
        color3 = "black"
        color4 = "white"
        color5 = "gray"
        colorbg = "lightgray"
    
    train = df.copy()
    
    sns.set_style("whitegrid")
         

    train[time_unit_name] = train["age_at_collection"]//time_unit_size

    plt.rc('axes', labelsize= 14)    # fontsize of the x and y labels
    plt.rc('xtick', labelsize= 8)
    plt.rc('legend', fontsize= 13)
    num_ticks = len(train[time_unit_name].unique())//2+1

    fig, axs = plt.subplots(len(important_features), 4, figsize=(40,len(important_features)*7))
    
    train_bacteria_name = train[[f"{bacteria_name}", time_unit_name]].groupby(time_unit_name).agg(np.mean).reset_index().sort_values(by=time_unit_name)
    months              = train_bacteria_name[time_unit_name].values
    total_num_of_crossings1, total_num_of_crossings2 = 0, 0
    for i in range(len(important_features)):
        
        name = important_features[i]
        
        # skip if numerator and denominator are same bacteria
        if name != bacteria_name:
            
            train_name          = train[[f"{name}", time_unit_name]].groupby(time_unit_name).agg(np.mean).reset_index()
            
            diff_ref_and_taxa1   = train_bacteria_name[f"{bacteria_name}"].values.reshape(-1) - train_name[f"{name}"].values.reshape(-1)
            diff_ref_and_taxa1   = -diff_ref_and_taxa1 if diff_ref_and_taxa1[0]<0 else diff_ref_and_taxa1
            signchange1 =  ((np.roll(np.sign(diff_ref_and_taxa1), 1) - np.sign(diff_ref_and_taxa1)) != 0).astype(int) 
            signchange1[0] = 0
            idx_cross1 = np.where(signchange1==1)[0]

            # plot abundances
            sns.pointplot(x=train[time_unit_name], y=train[f"{name}"].values, capsize=.2, alpha=0.4, ax=axs[i][0], color=color1, label=nice_name(name), markers="D")
            sns.pointplot(x=train[time_unit_name], y=train[f"{bacteria_name}"].values, capsize=.2, alpha=0.4, ax=axs[i][0], color=color2, label=f"{nice_name(bacteria_name)} (reference)", markers="*", size=220)
            axs[i][0].plot(months[idx_cross1], train_name[f"{name}"].values.reshape(-1)[idx_cross1], marker="X", markersize=15, color=color3, lw=0, alpha=1.0, label="crossed")
            axs[i][0].set_ylabel("Abundance")
            axs[i][0].set_xlabel(f"Age at collection [{time_unit_name}]")
            axs[i][0].set(facecolor=colorbg)
            axs[i][0].xaxis.set_major_locator(plt.MaxNLocator(num_ticks))
            
            x = np.arange(0, max(months)+1)
            p1, _ = np.polyfit(months, train_name[f"{name}"].values, 5, cov=True)    
            y1 = np.polyval(p1, x)
            y1[(y1==0.0)|(y1<1e-10)] = 1e-10
            p2, _ = np.polyfit(months, train_bacteria_name[f"{bacteria_name}"].values, 5, cov=True)    
            y2 = np.polyval(p2, x)
            y2[(y2==0.0)|(y2<1e-10)] = 1e-10

            diff_ref_and_taxa2   = y2 - y1
            diff_ref_and_taxa2   = -diff_ref_and_taxa2 if diff_ref_and_taxa2[0]<0 else diff_ref_and_taxa2
            signchange2 =  ((np.roll(np.sign(diff_ref_and_taxa2), 1) - np.sign(diff_ref_and_taxa2)) != 0).astype(int) 
            signchange2[0] = 0
            idx_cross2 = np.where(signchange2==1)[0]
            
            # plot smoothed abundances
            sns.pointplot(x=x, y=y1, capsize=.2, alpha=0.4, ax=axs[i][1], color=color1, label=nice_name(name), markers="D")
            sns.pointplot(x=x, y=y2, capsize=.2, alpha=0.4, ax=axs[i][1], color=color2, label=f"{nice_name(bacteria_name)} (reference)", markers="*", size=220)
            axs[i][1].plot(x[idx_cross2]-0.5, y1[idx_cross2], marker="X", markersize=15, color=color3, lw=0, alpha=1.0, label="crossed")
            axs[i][1].set_ylabel("Abundance")
            axs[i][1].set_xlabel(f"Age at collection [{time_unit_name}]")
            axs[i][1].set(facecolor=colorbg)
            axs[i][1].xaxis.set_major_locator(plt.MaxNLocator(num_ticks))
            
            # plot log-raios of the 2 bacteria abundances
            sns.pointplot(x=train[time_unit_name], y=[math.log2(x/y) for (x,y) in zip(train[f"{name}"].values, train[f"{bacteria_name}"].values)], capsize=.2, alpha=0.4, ax=axs[i][2], color=color5)
            axs[i][2].plot(np.linspace(0, 38, 10), np.zeros(10), lw=10, alpha=0.5, color=color4)
            axs[i][2].plot(months[idx_cross1]-0.5, np.zeros(len(idx_cross1)), alpha=0.7, marker="X", markersize=15, color=color3, lw=0)
            axs[i][2].set_ylabel("Log-Ratio")
            axs[i][2].set_xlabel(f"Age at collection [{time_unit_name}]")
            axs[i][2].set(facecolor=colorbg)
            axs[i][2].xaxis.set_major_locator(plt.MaxNLocator(num_ticks))
            
            # plot smoothed log-ratio of the 2 bacteria abundances
            #print([(k, j) for (k,j) in zip(y1, y2)])
            sns.pointplot(x=x, y=[math.log2(k/j) for (k,j) in zip(y1, y2)], capsize=.2, alpha=0.4, ax=axs[i][3], color=color5)
            axs[i][3].plot(np.linspace(0, 38, 10), np.zeros(10), lw=10, alpha=0.5, color=color4)
            axs[i][3].plot(x[idx_cross2]-0.5, np.zeros(len(idx_cross2)), alpha=0.7, marker="X", markersize=15, color=color3, lw=0)
            axs[i][3].set_ylabel("Log-Ratio")
            axs[i][3].set_xlabel(f"Age at collection [{time_unit_name}]")
            axs[i][3].set(facecolor=colorbg)
            axs[i][3].xaxis.set_major_locator(plt.MaxNLocator(num_ticks))

            custom_lines = [Line2D([0], [0], color=color1, ls="-", marker="o", lw=4),
                            Line2D([0], [0], color=color2, ls="-", marker="o", lw=4),
                            Line2D([0], [0], color=color3, marker="X", markersize=10, ls="--", lw=0),
                            Line2D([0], [0], color=color4, ls="-", alpha=0.4, lw=10)]
            axs[i][3].legend(custom_lines, [nice_name(important_features[i]), 
                                            f"{nice_name(bacteria_name)} (reference)", 
                                            "crossed", 
                                            "zero log-ratio"], loc="upper left", bbox_to_anchor=(1, 1))

            total_num_of_crossings1 += len(idx_cross1)
            total_num_of_crossings2 += len(idx_cross2)
   
    train[f"abundance_{bacteria_name}"] = train[bacteria_name].copy()
    for c in set(important_features):
        if c != bacteria_name:
            train[f"abundance_{c}"] = train[c].copy()
            train[c] = train.apply(lambda row: math.log2(row[f"abundance_{c}"]/row[f"abundance_{bacteria_name}"]), axis=1)
            
    _X_train, _y_train = df2vectors(train, important_features)
    rfr = Regressor(**parameters)
    gkf = list(GroupKFold(n_splits=n_splits).split(_X_train, _y_train, groups=train.subjectID.values))
    search = GridSearchCV(rfr, param_grid, cv=gkf)
    search.fit(_X_train, _y_train)
    estimator = search.best_estimator_
    _y_train_pred = estimator.predict(_X_train)
    _y_train = np.array(_y_train.values)/time_unit_size
    _y_train_pred = np.array(_y_train_pred)/time_unit_size
    mae   = round(np.mean(abs(_y_train_pred - _y_train)), 2)
    r2    = r2_score(_y_train, _y_train_pred) 
    
    pathlib.Path("/".join(file_name.split("/")[:-1])).mkdir(parents=True, exist_ok=True)
    plt.savefig(file_name)
    plt.tight_layout()
    if plot:
        plt.show()
    plt.clf()
    plt.close()

    return total_num_of_crossings1, total_num_of_crossings2, mae, r2


from sklearn.neighbors import LocalOutlierFactor

def update_reference_group_with_novelty_detection(df_all, feature_columns, local_outlier_factor_settings=dict(metric='braycurtis', n_neighbors=2)):
    """
    Features with bacteria only and Bray-Curtis distance as a metric.

    The number of neighbors considered (parameter n_neighbors) is typically set 1) greater than the minimum number of samples a cluster has to contain, so that other samples can be local outliers relative to this cluster, and 2) smaller than the maximum number of close by samples that can potentially be local outliers. In practice, such informations are generally not available, so we take it to be 5. Information on other algorithms used: https://scikit-learn.org/stable/modules/outlier_detection.html#novelty-with-lof
    """
    df = df_all.copy()
    
    X_train = df[df["reference_group"]==True][feature_columns].values

    lof = LocalOutlierFactor(novelty=True, **local_outlier_factor_settings)
    lof.fit(X_train)
    
    X_test = df[df["reference_group"]==False][feature_columns].values
    y_test = lof.predict(X_test)
    
    df.loc[df["reference_group"]==False, "reference_group"] = y_test==1
    
    return df["reference_group"].values


from plotly.subplots import make_subplots
import plotly.graph_objects as go

def gridsearch_novelty_detection_parameters(df_all, parameter_name, parameter_vals, feature_columns, metadata_columns, meta_and_feature_columns, num_runs=5, website=False, layout_settings=None):
    df = df_all.copy()
    
    df_stats = pd.DataFrame(data={"parameter":[], "run":[], "num_ref":[], "num_nonref":[], "accuracy":[], "columns_type":[]})

    # after modification
    df["reference_group"].value_counts()
    
    d = {}
    for n in parameter_vals:
        d["parameter"] = n

        settings = dict(metric='braycurtis')
        settings[parameter_name] = n

        # temp: reference_group1
        df = df.assign(reference_group1=update_reference_group_with_novelty_detection(df, meta_and_feature_columns, local_outlier_factor_settings=settings))

        # after modification
        d["num_ref"] = df["reference_group1"].value_counts()[True]
        d["num_nonref"] = df["reference_group1"].value_counts()[False]

        for i in range(num_runs):
            d["run"] = i
            output1 = two_groups_analysis(df, feature_columns, references_we_compare="reference_group1", nice_name=lambda x: x, style="dot", show=False, website=False);
            d["columns_type"] = "taxa"
            d["accuracy"] = output1["accuracy"]
            df_stats = df_stats.append(d, ignore_index=True)
            output2 = two_groups_analysis(df, metadata_columns, references_we_compare="reference_group1", nice_name=lambda x: x, style="dot", show=False, website=False);
            d["columns_type"] = "meta"
            d["accuracy"] = output2["accuracy"]
            df_stats = df_stats.append(d, ignore_index=True)
            output3 = two_groups_analysis(df, meta_and_feature_columns, references_we_compare="reference_group1", nice_name=lambda x: x, style="dot", show=False, website=False);
            d["columns_type"] = "meta and taxa"
            d["accuracy"] = output3["accuracy"]
            df_stats = df_stats.append(d, ignore_index=True)

    num_cols = 3
    fig = make_subplots(rows=1, cols=num_cols, horizontal_spacing=0.1, subplot_titles=df_stats.columns_type.unique())

    layout_settings_default = dict(
        height=400, 
        width=1100, 
        paper_bgcolor="white",
        plot_bgcolor='rgba(0,0,0,0)', 
        margin=dict(l=0, r=0, b=0, pad=0),
        title_text="Novelty Algorithm Performance measure for different value of parameter",
        font=dict(size=10),
        yaxis=dict(position=0.0)
    )

    if layout_settings is None:
        layout_settings = {}
    layout_settings_final = {**layout_settings_default, **layout_settings}
    
    for idx, ct in enumerate(df_stats.columns_type.unique()):
        df_stats1 = df_stats[df_stats.columns_type==ct]
        fig.add_trace(go.Scatter(
                x=df_stats1.groupby(by="parameter").agg(np.median)["accuracy"].index,
                y=df_stats1.groupby(by="parameter").agg(np.median)["accuracy"],
                error_y=dict(
                    type='data', # value of error bar given in data coordinates
                    array=df_stats1.groupby(by="parameter").agg(np.std)["accuracy"],
                    visible=True), name=ct
            ), row=1, col=idx%num_cols+1)
        fig.update_xaxes(title="parameter", row=1, col=idx%num_cols+1)  # gridcolor='lightgrey'
        fig.update_yaxes(title="accuracy", row=1, col=idx%num_cols+1)  # gridcolor='lightgrey'

    fig.update_layout()
    if not website:
        fig.show()
        
    return df_stats, fig


import matplotlib
matplotlib.use('agg')
import numpy as np
import pandas as pd
from scipy.stats import ttest_ind, sem
import seaborn as sns
import matplotlib.pyplot as plt
import plotly.graph_objects as go
from scipy.stats import ttest_rel, wilcoxon, ttest_ind, spearmanr, pearsonr, t
from skbio.stats.composition import multiplicative_replacement
from scipy.stats import ttest_ind
import plotly.express as px


def plot_abundances_and_log(b1, b2, rmp_data, nice_name):
    
    fig, axs = plt.subplots(3, 1, figsize=(5, 13))
    sns.pointplot(x="label", y='bacteria1', hue="time", data=rmp_data, ax=axs[0])
    axs[0].set_ylabel(nice_name(b1))
    sns.pointplot(x="label", y='bacteria2', hue="time", data=rmp_data, ax=axs[1])
    axs[1].set_ylabel(nice_name(b2))
    sns.pointplot(x="label", y='log(bacteria1/bacteria2)', hue="time", data=rmp_data, ax=axs[2])
    axs[2].set_ylabel(r'$\log( \frac{' + nice_name(b1) + '}{' + nice_name(b2) + '})$')
    axs[1].get_legend().remove();axs[2].get_legend().remove()
    axs[0].set_xlabel("");axs[1].set_xlabel("");axs[2].set_xlabel("");
    
    plt.show()
    

def plot_trajectory(ax, df, name, title='', left_title='', ypad=0, logscale=False, norm1=False):

    subs = df['time'].value_counts().index
    
    for i, sub in enumerate(np.sort(subs)):
        subdf = df.loc[df['time'] == sub]
        subdf = subdf.sort_values(by='label').iloc[::-1]

        x = np.arange(2)
        y = subdf.groupby('label').mean()[name][::-1]
        
        if norm1:
            y = y / y[0]
        else:
            y = y - y[0]
        
        e = subdf.groupby('label').agg(sem)[name][::-1]
        ax.errorbar(x, y, yerr=e, label=sub, marker='o')

    ax.set_xticks([0, 1])            
    ax.set_xticklabels(y.index)

    if logscale:
        ax.set_yscale('log')

    ax.set_title(title, fontsize=14)
    ax.set_xlabel('')
    for x in ax.get_xticklabels():
        x.set_fontsize(12)

    ax.set_ylabel(left_title, fontsize=14, labelpad=ypad, rotation=0)

    for y in ax.get_yticklabels():
        y.set_fontsize(12)   
        

def plot_abundance_and_ratio(b1, b2, table, metadata, nice_name):
    # TODO: make time column general, to support
    rmp_table = table.apply(lambda x: x * 100 / x.sum(), axis=1)
    rmp_table = pd.DataFrame(data=multiplicative_replacement(rmp_table.values), columns=rmp_table.columns, index=rmp_table.index)
    
    print(f"Bacteria 1: {b1}\nBacteria 2: {b2}")

    rmp_data = pd.DataFrame({
        'log(bacteria1/bacteria2)': np.log10(rmp_table[b1]) - np.log10(rmp_table[b2]),
        'bacteria1': rmp_table[b1],
        'bacteria2': rmp_table[b2],
        'label': metadata['label'],
        'time': metadata['time']    
    })

    plot_abundances_and_log(b1, b2, rmp_data, nice_name)

    num_taxa = 'bacteria1'
    denom_taxa = 'bacteria2'
    ypad_scalar = 4

    fig, axes = plt.subplots(3, 1, figsize=(3, 8))
    plot_trajectory(axes[0], rmp_data, num_taxa, 
                    logscale=False, left_title=r"$\it{%s}$" % nice_name(b1), ypad=len(nice_name(b1))*ypad_scalar)
    plot_trajectory(axes[1], rmp_data, denom_taxa,
                    logscale=False, left_title=r"$\it{%s}$" % nice_name(b2), ypad=len(nice_name(b2))*ypad_scalar)
    plot_trajectory(axes[2], rmp_data, 'log(' + num_taxa + '/' + denom_taxa + ')', 
                    left_title=r'$\log( \frac{' + nice_name(b1) + '}{' + nice_name(b2) + '})$', ypad=max(len(nice_name(b1)), len(nice_name(b1)))*ypad_scalar)

    axes[0].legend(loc='upper center', bbox_to_anchor=(0.5, 1.5), ncol=1, fancybox=True, shadow=True)
    axes[2].set_xlabel('% rel')

    # Test if bacteria (numerator, denominator, log-ratio) is changing in the data?
    # numerator taxa - bacteria 1
    subrmp_data = rmp_data.dropna()
    statistic, pval = ttest_ind(np.array(subrmp_data.loc[subrmp_data['label'] == 'reference', num_taxa]),
                                np.array(subrmp_data.loc[subrmp_data['label'] == 'other', num_taxa]))
    print("statistic =", statistic, "\np-value =",pval)
    print("The means are NOT significantly different." if pval>0.05 else "The means are significantly different.")

    # denominator taxa - bacteria 2
    subrmp_data = rmp_data.dropna()
    statistic, pval = ttest_ind(np.array(subrmp_data.loc[subrmp_data['label'] == 'reference', denom_taxa]),
                                np.array(subrmp_data.loc[subrmp_data['label'] == 'other', denom_taxa]))
    print("statistic =", statistic, "\np-value =",pval)
    print("The means are NOT significantly different." if pval>0.05 else "The means are significantly different.")

    # log ration of two taxa
    subrmp_data = rmp_data.dropna()
    statistic, pval = ttest_ind(
        np.array(subrmp_data.loc[subrmp_data['label'] == 'reference', 
                              'log(' + num_taxa + '/' + denom_taxa + ')']),
        np.array(subrmp_data.loc[subrmp_data['label'] == 'other', 
                              'log(' + num_taxa + '/' + denom_taxa + ')']))
    print("statistic =", statistic, "\np-value =",pval)
    print("The means are NOT significantly different." if pval>0.05 else "The means are significantly different.")
    
def differential_ranking_with_hybrid_matplotlib(beta, df_hybrid_and_all, reference_column):
    _df = pd.merge(df_hybrid_and_all.set_index("bacteria_name"), beta, left_index=True, right_index=True)
    _df = _df.sort_values(by=reference_column)
    x = np.arange(_df.shape[0])

    fig = plt.figure(figsize=(25, 10))
    
    for b in _df.hybrid_bacteria_name.unique():
        idx = np.where(_df.hybrid_bacteria_name==b)[0]
        color = _df.iloc[idx].drop_duplicates().color.values[0]
        plt.bar(x[idx], _df.iloc[idx][reference_column].values, 0.8, alpha=0.2 if len(b)==0 else 0.8, edgecolor=color, facecolor=color, lw=1, label="%s"%b[0:].replace("_", " ") if len(b)!=0 else None)  #label="%s" % genus[3:].replace("_", " "),
    
    plt.xlabel('species');plt.xticks([]);plt.ylabel(r'$\log (\frac{referemce}{others}) + K$', fontsize=20, labelpad=90, rotation=0);
    plt.tick_params(axis='both', which='major', labelsize=14)
    plt.plot(x, np.array([0]*len(x)), c='k', lw=1)

    plt.legend(fontsize=14)
    plt.tight_layout()
    plt.show()
    
def differential_ranking_with_hybrid_big_matplotlib(beta, df_hybrid_and_all, reference_column):
    _df = pd.merge(df_hybrid_and_all.set_index("bacteria_name"), beta, left_index=True, right_index=True)
    _df = _df.sort_values(by=reference_column)

    # Plot the figure.
    plt.figure(figsize=(30, 100))
    plt.rc('ytick', labelsize=15)

    ax =  _df[reference_column].plot(kind='barh', color=_df.color)
    
    rects = ax.patches

    # For each bar: Place a label
    for rect in rects:
        # Get X and Y placement of label from rect.
        x_value = rect.get_width()
        y_value = rect.get_y() + rect.get_height() / 2

        # Number of points between bar and label. Change to your liking.
        space = 5
        # Vertical alignment for positive values
        ha = 'left'

        # If value of bar is negative: Place label left of bar
        if x_value < 0:
            # Invert space to place label to the left
            space *= -1
            # Horizontally align label at right
            ha = 'right'

        # Create annotation
        plt.annotate(
            f"{x_value:.2e}",                      # Use `label` as label
            (x_value, y_value),         # Place label at end of the bar
            xytext=(space, 0),          # Horizontally shift label by `space`
            textcoords="offset points", # Interpret `xytext` as offset in points
            va='center',                # Vertically center label
            ha=ha,
            size=15)                      # Horizontally align label differently for
                                        # positive and negative values.

    ax.set_xlabel(r'$\log (\frac{referemce}{others}) + K$')
    ax.set_ylabel('Bacteria') 
    ax.set_yticklabels(_df.index)
    
    plt.show()

    
def differential_ranking_with_hybrid(beta, df_hybrid_and_all, reference_column, height=600, width=1600, file_name=None):
    # reference_column="label[T.reference]"
    _df = pd.merge(df_hybrid_and_all.set_index("bacteria_name"), beta, left_index=True, right_index=True)
    _df = _df.sort_values(by=reference_column)
    x = np.arange(_df.shape[0])
    
    fig = go.Figure()    
    fig.add_trace(go.Bar(x=x, y=_df[reference_column].values, marker_color=_df.color,
                 customdata=[(x, y) for x, y in zip(_df.index, _df.hybrid_bacteria_name)],
                 hovertemplate = '<b>Taxa: %{customdata[0]}</b><br><br>'+
                                '<b>Log-ratio</b>: %{y}<br>'+
                                '<b>Hybrid class</b>: %{customdata[1]}<br>'))
    
    fig.update_xaxes(title="Species", gridcolor='lightgrey', ticks="", showticklabels=False) 
    fig.update_yaxes(title=r'$\log (\frac{referemce}{others}) + K$', gridcolor='lightgrey')  


    fig.update_layout(height=height, width=width, 
                      plot_bgcolor='rgba(0,0,0,0)', 
                      margin=dict(l=0, r=0, b=0, pad=0),
                      title_text="Songbird's Multinomial Regression")
    if file_name:
        fig.write_html(file_name)

    fig.show()


# country colors
from sklearn.ensemble import RandomForestRegressor
from catboost import CatBoostRegressor

col_rus = "#E3211B"
col_est = "#0095C8"
col_fin = "#00A24A"
alpha   = 0.7

Regressor = CatBoostRegressor
parameters = {"loss_function": "MAE",
              "random_state": 42,
              "allow_writing_files": True,
             "verbose":False}
param_grid = { 'learning_rate': [0.5, 0.1],
               #"depth": [5, 10],
               #"iterations": [100, 500, 1000]
             }


import io
import base64
import html
from shap.plots._force_matplotlib import draw_additive_plot
import shap
import pandas as pd


def df2vectors(_df, feature_cols=None):
    """Function that converts data frame to corresponding feature vector X and the label vector y
    
    This function is mostly used to convert pandas dataframe objects to lists, and then propagated to
    different ML methods (i.e. CatBoost, RandomForest) in order to perform the prediction.

    Parameters
    ----------
    _df: pd.DataFrame
        Data frame that we want to convert to a feature vector X with its corresponding labels y.
    feature_cols: list
        Column names that represent the features that will be used in a feature vector X (e.g. bacteria names, and/or meta data). There
        should not be any column that is ID (e.g. sample id, etc.)
    
    Returns
    -------
    X: list of lists
        Feature matrix with dimension (n_samples, n_features). 
    y: list
        Label vector with dimension (n_samples,)
    """
    
    # labels, what we want to predict -> age of the infant
    y = _df["age_at_collection"]
    
    _df = _df[feature_cols]
    
    # features, what is given to us -> bacteria info
    X = _df.values

    return X, y


def get_bacteria_names(df, bacteria_fun=lambda x: x.startswith("k__")):
    return list(df.columns[df.columns.map(bacteria_fun)])


import shap
from sklearn.ensemble import RandomForestClassifier
import seaborn as sns
import matplotlib.pyplot as plt
from plotly.tools import mpl_to_plotly
import numpy as np
from sklearn.model_selection import GroupShuffleSplit
from sklearn.metrics import plot_confusion_matrix
from sklearn.metrics import confusion_matrix
import itertools
from io import BytesIO

def fig_to_uri(in_fig, close_all=True, **save_args):
    # type: (plt.Figure) -> str
    """
    Save a figure as a URI
    :param in_fig:
    :return:
    """
    out_img = BytesIO()
    in_fig.savefig(out_img, format='png', **save_args)
    if close_all:
        in_fig.clf()
        plt.close('all')
    out_img.seek(0)  # rewind file
    encoded = base64.b64encode(out_img.read()).decode("ascii").replace("\n", "")
    return "data:image/png;base64,{}".format(encoded)

def _plot_confusion_matrix(cm, title, classes=['False', 'True'],
                          cmap=plt.cm.Blues, save=False, saveas="MyFigure.png", website=False):
    
    # print Confusion matrix with blue gradient colours
    plt.rcParams.update({'font.size': 15})
    
    cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
    
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=90)
    plt.yticks(tick_marks, classes)

    fmt = '.1%'
    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, format(cm[i, j], fmt),
                 horizontalalignment="center",
                 color="white" if cm[i, j] > thresh else "black")

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')
    
    if save:
        plt.savefig(saveas, dpi=100)

    if not website:
        plt.show()
    else:
        return fig_to_uri(plt)

def two_groups_analysis(df_all, feature_cols, references_we_compare, test_size=0.5, n_splits=5, nice_name=lambda x: x, style="dot", show=False, website=True, layout_height=1000, layout_width=1000, max_display=20):
    """Style can be dot or hist"""
    if show: 
        print("two_groups_analysis")
        shap.initjs() 
    
    df = df_all.copy()
    
    df1 = df[df[references_we_compare]==True]
    df2 = df[df[references_we_compare]==False]
    
    min_samples = df[references_we_compare].value_counts(sort=True).values[-1]
    min_group   = df[references_we_compare].value_counts(sort=True).index[-1]
    
    #print(min_samples/len(df))
    if min_group==True:
        df2 = df2.sample(n=int(min_samples))  #+(len(df)-min_samples)*0.1)
    else:
        df1 = df1.sample(n=int(min_samples))
    
    df = pd.concat([df1, df2])

    df["dataset_type_classification"] = ""
    if len(df[df[references_we_compare]==True])>0:
        train_idx1, test_idx1 = next(GroupShuffleSplit(test_size=test_size, n_splits=n_splits, random_state=7).split(df[df[references_we_compare]==True].index, groups=df[df[references_we_compare]==True]['subjectID']))
        df.loc[df[df[references_we_compare]==True].iloc[train_idx1].index, "classification_dataset_type"] = "Train-1"
        df.loc[df[df[references_we_compare]==True].iloc[test_idx1].index, "classification_dataset_type"] = "Test-1"

    if len(df[df[references_we_compare]==False])>0:
        train_idx2, test_idx2 = next(GroupShuffleSplit(test_size=test_size, n_splits=n_splits, random_state=7).split(df[df[references_we_compare]==False].index, groups=df[df[references_we_compare]==False]['subjectID']))
        df.loc[df[df[references_we_compare]==False].iloc[train_idx2].index, "classification_dataset_type"] = "Train-2"
        df.loc[df[df[references_we_compare]==False].iloc[test_idx2].index, "classification_dataset_type"] = "Test-2"

    #print(df.classification_dataset_type.value_counts())
    df.loc[df[references_we_compare]==True, "classification_label"] = 1
    df.loc[df[references_we_compare]==False, "classification_label"] = -1

    df_train = df[df.classification_dataset_type.str.startswith("Train")]
    df_test = df[df.classification_dataset_type.str.startswith("Test")]
    
    X_train = df_train[feature_cols]
    y_train = df_train.classification_label.values.astype('int')
    X_test = df_test[feature_cols]
    y_test = df_test.classification_label
    
    m = RandomForestClassifier(n_estimators=140, random_state=0, max_samples=0.8)
    m.fit(X_train, y_train)
    
    fig, ax = plt.subplots()
    
    explainer = shap.TreeExplainer(m)
    shap_values = explainer.shap_values(X_train)
    
    feature_names = list(map(nice_name, X_train.columns))

    y_test_pred = m.predict(X_test)
    print(type(y_test_pred))
    print(type(y_test))
    if isinstance(y_test, pd.Series):
        y_test = np.array(y_test.values, dtype=type(y_test_pred[0]))

    print(type(y_test_pred))
    print(type(y_test))
    print(type(y_test_pred[0]))
    print(type(y_test[0]))

    cm_test = confusion_matrix(y_test_pred, y_test)
    acc = 100*(cm_test[0][0]+cm_test[1][1]) / (sum(cm_test[0]) + sum(cm_test[1]))

    max_limit = max(20, min(max_display, len(feature_names)))

    if not website:
        sns.set_style("whitegrid")
        
        if style == "dot":
            shap.summary_plot(shap_values[0], features=X_train, feature_names=feature_names, show=show, max_display=max_display)
        elif style == "hist":
            shap.summary_plot(shap_values, features=X_train, feature_names=feature_names, class_names=["reference", "other"], show=show, max_display=max_display)


        
        if show:
            plot_confusion_matrix(m, X_test, y_test)  
            plt.show() 
            _plot_confusion_matrix(cm_test,"Confusion matrix", ['False', 'True'])

            print('Total OTHER detected in test set: ' + str(cm_test[1][1]) + ' / ' + str(cm_test[1][1]+cm_test[1][0]))
            print('Total REFERENCE transactions detected in test set: ' + str(cm_test[0][0]) + ' / ' + str(cm_test[0][1]+cm_test[0][0]))
            print('Probability to detect a OTHER in the test set: ' + str(cm_test[1][1]/(cm_test[1][1]+cm_test[1][0])))
            print('Probability to detect a REFERENCE in the test set: ' + str(cm_test[0][0]/(cm_test[0][1]+cm_test[0][0])))
            print("Accuracy of unsupervised anomaly detection model on the test set: "+f"{acc:.2f}" + "%")
            print("-----\n\n")

        output = dict(top_features_list=list(X_train.columns[np.argsort(np.abs(shap_values[0]).mean(0))])[::-1],
                      accuracy=acc)

        return output
    else:
        if style == "dot":
            shap.summary_plot(shap_values[0], features=X_train, show=show, max_display=max_display)

            fig =  mpl_to_plotly(fig)
            
            fig.update_xaxes(title="SHAP value (impact on model output)", 
                            showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray') 
            fig.update_yaxes(title="Features", 
                             tickmode='array',
                             tickvals=list(range(0, max_limit)),  #list(X_train.columns[np.argsort(np.abs(shap_values[0]).mean(0))])[::-1][:min(20, len(feature_names))], # 
                             ticktext=list(map(nice_name, list(X_train.columns[np.argsort(np.abs(shap_values[0]).mean(0))])[::-1][:max_limit][::-1])),
                            showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray')  
            fig.update_layout(height=layout_height, width=layout_width,
                            #paper_bgcolor="white",#'rgba(0,0,0,0)', 
                            plot_bgcolor='rgba(0,0,0,0)', 
                            margin=dict(l=0, r=0, b=0, pad=0),
                            title_text="Classification Important Features")
            
            #fig.update_traces(marker=dict(color="red"), selector=dict(type="scatter", mode="marker"))
#             fig.for_each_trace(
#                 lambda trace: trace.update(marker_symbol="square") if trace.name == "trace 39" else (),
#             )
            img_src = _plot_confusion_matrix(cm_test,"Confusion matrix", ['False', 'True'], website=website)

            return fig, img_src

        elif style == "hist":
            shap.summary_plot(shap_values, features=X_train, class_names=["reference", "other"], show=show, max_display=max_display)
            
            fig =  mpl_to_plotly(fig)
            
            fig.update_xaxes(title="SHAP value (impact on model output)", 
                            showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray') 
            fig.update_yaxes(title="Features", 
                             tickmode='array',
                             tickvals=list(range(0, max_limit)),  #list(X_train.columns[np.argsort(np.abs(shap_values[0]).mean(0))])[::-1][:min(20, len(feature_names))], # 
                             ticktext=list(map(nice_name, list(X_train.columns[np.argsort(np.abs(shap_values[0]).mean(0))])[::-1][:max_limit][::-1])),
                            showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray')  
            fig.update_layout(height=layout_height, width=layout_width,
                            #paper_bgcolor="white",#'rgba(0,0,0,0)', 
                            plot_bgcolor='rgba(0,0,0,0)', 
                            margin=dict(l=0, r=0, b=0, pad=0),
                            title_text="Classification Important Features")
            return fig


import matplotlib
matplotlib.use('agg')
from pickle import FALSE
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from matplotlib.lines import Line2D
import seaborn as sns
from catboost import Pool
from sklearn.model_selection import cross_val_score,cross_val_predict
from sklearn.model_selection import GroupKFold
from sklearn.metrics import r2_score
from sklearn.feature_selection import VarianceThreshold
import statsmodels.api as sm
import scipy.stats as stats
import scipy as sp
import pathlib
import shap
from microbiome.variables import *
from microbiome.helpers import df2vectors
from microbiome.statistical_analysis import *
import plotly.graph_objects as go
import plotly.express as px
from itertools import combinations 
from plotly.subplots import make_subplots
from sklearn.model_selection import GridSearchCV


def topK_important_features(k, estimator, df, feature_cols, n_splits, estimator_for_fit, save=False, file_name=None):
    """Get top k important features for the estimator

    k: int
        Number of important features we want to have in the model.
    estimator:
        Model for the trajectory.
    df: pd.DataFrame
        Dataset on which we want to find the top important features.
    feature_cols: list
        List of bacteria for the estimator.
    n_splits: int
        Number of split for the group cross validation.
    estimator_for_fit:
        Object to use to fit the data (same as the estimator)
    save: bool
        If True, the search for the important features will be performed and sabed in the end. Otherwise, it will read the important features from some 
        of the previous runs from a file.
    file_name: str
        Name of the file to save data in or read from.

    Returns
    -------
    important_features: list
        List of important features.
    """
    if file_name is None:
        raise Exception("You should specify a file you wish to save the list of important features")

    # f'OUTPUT_FILES/important_features_{DATASET_LEVEL}_{SAVE_IDENTIFIER}.txt'
    if save:
        X, y = df2vectors(df, feature_cols)
        
        shap_values = estimator.get_feature_importance(Pool(X, y), type="ShapValues")
        feature_importance = pd.DataFrame(list(zip(feature_cols, np.abs(shap_values).mean(0))), columns=['bacteria_name','feature_importance_vals'])
        feature_importance.sort_values(by=['feature_importance_vals'], ascending=False, inplace=True)
        important_features = list(feature_importance.bacteria_name.values)

        features_num = []
        accuracy_means = []
        accuracy_stds = []
        maes = []
        r2s = []

        i = 1
        while i <  len(important_features[:k])+1:
            done = 0
            X, y = df2vectors(df, important_features[:i])
            while done < 3:
                try:
                    scores = cross_val_score(estimator_for_fit, X, y, groups=df.subjectID.values, cv=GroupKFold(n_splits), verbose=0)
                    y_pred = cross_val_predict(estimator_for_fit, X, y, groups=df.subjectID.values, cv=GroupKFold(n_splits), verbose=0)
                    accuracy_means.append(scores.mean())
                    accuracy_stds.append(scores.std())
                    maes.append(np.mean(abs(y_pred - y)))
                    r2s.append(r2_score(y, y_pred))
                    features_num.append(i)
                    i+=2
                    done = 4
                except Exception as e:
                    print(e)
                    done += 1


        fig = make_subplots(rows=1, cols=3)
        
        i = np.argmin(maes)
        fig.add_trace(go.Scatter(
                x=features_num,
                y=maes,
                name="MAE",
                hovertemplate =None),
                row=1, col=1)
        fig.add_trace(go.Scatter(
                x=[features_num[i]],
                y=[maes[i]],
                mode="markers",
                marker_symbol="star",
                marker_size=15,
                marker_color="green",
                name="optimal MAE",
                showlegend=False,
                hovertemplate =None),
                row=1, col=1)
        fig.update_xaxes(title="Number of important features", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=1) 
        fig.update_yaxes(title="MAEs", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=1)
                
        j = np.argmax(accuracy_means)
        fig.add_trace(go.Scatter(
                x=features_num,
                y=accuracy_means,
                name="Accuracy",
                hovertemplate = None,
                error_y=dict(
                        type='data', 
                        array=accuracy_stds,
                        visible=True)),
                row=1, col=2)
        fig.add_trace(go.Scatter(
                x=[features_num[i]],
                y=[accuracy_means[i]],
                mode="markers",
                marker_symbol="star",
                marker_size=15,
                marker_color="green",
                name="optimal MAE",
                showlegend=False,
                hovertemplate =None),
                row=1, col=2)
        fig.update_xaxes(title="Number of important features", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=2) 
        fig.update_yaxes(title="Accuracy", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=2)
        
        fig.add_trace(go.Scatter(
                x=features_num,
                y=r2s,
                name="R-square",
                hovertemplate = None),
                row=1, col=3)
        fig.add_trace(go.Scatter(
                x=[features_num[i]],
                y=[r2s[i]],
                mode="markers",
                marker_symbol="star",
                marker_size=15,
                marker_color="green",
                name="optimal MAE",
                showlegend=True,
                hovertemplate =None),
                row=1, col=3)
        fig.update_xaxes(title="Number of important features", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=3)  
        fig.update_yaxes(title="R-squared", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=3)  

        fig.update_layout(height=500, width=1200, 
                        #paper_bgcolor="white",#'rgba(0,0,0,0)', 
                        plot_bgcolor='rgba(0,0,0,0)', 
                        margin=dict(l=0, r=0, b=0, pad=0),
                        title_text="Top Important Features",
                        hovermode="x")
        fig.show()
        
        # excract only top important features
        important_features = important_features[:features_num[i]]

        with open(file_name, 'w') as f:
            f.writelines(f"{feature}\n" for feature in important_features)

    else:
        important_features = []

        with open(file_name, 'r') as f:
            filecontents = f.readlines()

            for line in filecontents:
                important_features.append(line[:-1])
                
    return important_features

def remove_nzv(save, df, feature_cols, n_splits, estimator_for_fit, nzv_thresholds=None):
    feature_cols = np.array(feature_cols)
    """Remove features with near-zero-variance"""
    if save:
        accuracy_means = []
        accuracy_stds = []
        maes = []
        r2s = []
        features_num = []
        
        for threshold in nzv_thresholds:
            # filter near zero variance features
            constant_filter = VarianceThreshold(threshold=threshold)
            constant_filter.fit(df[feature_cols])
            idx = np.where(constant_filter.get_support())[0]
            constant_columns = [column for column in feature_cols if column not in feature_cols[idx]]
            features_num.append(len(feature_cols)-len([x for x in constant_columns if x.startswith('k__')] ))
            feature_cols_new = list(set(feature_cols) - set(constant_columns))

            X, y = df2vectors(df, feature_cols_new)

            scores = cross_val_score(estimator_for_fit, X, y, groups=df.subjectID.values, cv=GroupKFold(n_splits), verbose=0)
            accuracy_means.append(scores.mean())
            accuracy_stds.append(scores.std())

            y_pred = cross_val_predict(estimator_for_fit, X, y, groups=df.subjectID.values, cv=GroupKFold(n_splits), verbose=0)
            maes.append(np.mean(abs(y_pred - y)))
            r2s.append(r2_score(y, y_pred))

        fig = make_subplots(rows=1, cols=3)
        
        i = np.argmin(maes)
        fig.add_trace(go.Scatter(
                x=nzv_thresholds,
                y=maes,
                name="MAE",
                hovertemplate =None),
                row=1, col=1)
        fig.add_trace(go.Scatter(
                x=[nzv_thresholds[i]],
                y=[maes[i]],
                mode="markers",
                marker_symbol="star",
                marker_size=15,
                marker_color="green",
                name="optimal MAE",
                showlegend=False,
                hovertemplate =None),
                row=1, col=1)
        fig.update_xaxes(title="NZV thresholds", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=1) 
        fig.update_yaxes(title="MAEs", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=1) 
                
        j = np.argmax(accuracy_means)
        fig.add_trace(go.Scatter(
                x=nzv_thresholds,
                y=accuracy_means,
                name="Accuracy",
                hovertemplate = None,
                error_y=dict(
                        type='data', 
                        array=accuracy_stds,
                        visible=True)),
                row=1, col=2)
        fig.add_trace(go.Scatter(
                x=[nzv_thresholds[i]],
                y=[accuracy_means[i]],
                mode="markers",
                marker_symbol="star",
                marker_size=15,
                marker_color="green",
                name="optimal MAE",
                showlegend=False,
                hovertemplate =None),
                row=1, col=2)
        fig.update_xaxes(title="NZV thresholds", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=2) 
        fig.update_yaxes(title="Accuracy", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=2)  
        
        fig.add_trace(go.Scatter(
                x=nzv_thresholds,
                y=r2s,
                name="R-square",
                hovertemplate = None),
                row=1, col=3)
        fig.add_trace(go.Scatter(
                x=[nzv_thresholds[i]],
                y=[r2s[i]],
                mode="markers",
                marker_symbol="star",
                marker_size=15,
                marker_color="green",
                name="optimal MAE",
                showlegend=True,
                hovertemplate =None),
                row=1, col=3)
        fig.update_xaxes(title="NZV thresholds", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=3) 
        fig.update_yaxes(title="R-squared", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=3) 

        fig.update_layout(height=500, width=1200, 
                        #paper_bgcolor="white",
                        plot_bgcolor='rgba(0,0,0,0)', 
                        margin=dict(l=0, r=0, b=0, pad=0),
                        title_text="Near Zero Variance",
                        hovermode="x")
        fig.show()

        nzv_threshold = nzv_thresholds[j]
    else:
        # 26-Nov-2020
        nzv_threshold = 0.01
        
    constant_filter = VarianceThreshold(threshold=nzv_threshold)
    constant_filter.fit(df[feature_cols])
    feature_cols = np.array(feature_cols)
    constant_columns = [column for column in feature_cols if column not in feature_cols[np.where(constant_filter.get_support())[0]]]
    feature_cols_new = list(set(feature_cols) - set(constant_columns))
    print(f"Number of features left after removing features with variance {nzv_threshold} or smaller: {len(feature_cols_new)}/{len(feature_cols)}")

    return feature_cols_new

def remove_correlated(save, df, feature_cols, n_splits, estimator_for_fit, correlation_thresholds=None):
    """"Remove correlated values"""
    # the higher the number the less it is removing
    if save:
        #correlation_thresholds = [0.0, 0.1, 0.3, 0.8]
        accuracy_means = []
        accuracy_stds = []
        maes = []
        r2s = []
        features_num = []

        for threshold in correlation_thresholds:
            correlated_features = set()
            correlation_matrix = df[feature_cols].corr()
            for i in range(len(correlation_matrix.columns)):
                for j in range(i):
                    if abs(correlation_matrix.iloc[i, j]) > threshold:
                        colname = correlation_matrix.columns[i]
                        correlated_features.add(colname)
                        
            feature_cols_ncorr = list(set(feature_cols)-correlated_features)
            features_num.append(len(feature_cols_ncorr))

            X_healthy, y_healthy = df2vectors(df, feature_cols_ncorr)
            
            scores = cross_val_score(estimator_for_fit, X_healthy, y_healthy, groups=df.subjectID.values, cv=GroupKFold(n_splits), verbose=0)
            accuracy_means.append(scores.mean())
            accuracy_stds.append(scores.std())

            y_healthy_pred = cross_val_predict(estimator_for_fit, X_healthy, y_healthy, groups=df.subjectID.values, cv=GroupKFold(n_splits), verbose=0)
            maes.append(np.mean(abs(y_healthy_pred - y_healthy)))
            r2s.append(r2_score(y_healthy, y_healthy_pred))

        fig = make_subplots(rows=1, cols=3)
        
        i = np.argmin(maes)
        fig.add_trace(go.Scatter(
                x=correlation_thresholds,
                y=maes,
                name="MAE",
                hovertemplate =None),
                row=1, col=1)
        fig.add_trace(go.Scatter(
                x=[correlation_thresholds[i]],
                y=[maes[i]],
                mode="markers",
                marker_symbol="star",
                marker_size=15,
                marker_color="green",
                name="optimal MAE",
                showlegend=False,
                hovertemplate =None),
                row=1, col=1)
        fig.update_xaxes(title="NZV thresholds", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=1) 
        fig.update_yaxes(title="MAEs", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=1)  
                
        j = np.argmax(accuracy_means)
        fig.add_trace(go.Scatter(
                x=correlation_thresholds,
                y=accuracy_means,
                name="Accuracy",
                hovertemplate = None,
                error_y=dict(
                        type='data',
                        array=accuracy_stds,
                        visible=True)),
                row=1, col=2)
        fig.add_trace(go.Scatter(
                x=[correlation_thresholds[i]],
                y=[accuracy_means[i]],
                mode="markers",
                marker_symbol="star",
                marker_size=15,
                marker_color="green",
                name="optimal MAE",
                showlegend=False,
                hovertemplate =None),
                row=1, col=2)
        fig.update_xaxes(title="NZV thresholds", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=2) 
        fig.update_yaxes(title="Accuracy", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=2) 
        
        fig.add_trace(go.Scatter(
                x=correlation_thresholds,
                y=r2s,
                name="R-square",
                hovertemplate = None),
                row=1, col=3)
        fig.add_trace(go.Scatter(
                x=[correlation_thresholds[i]],
                y=[r2s[i]],
                mode="markers",
                marker_symbol="star",
                marker_size=15,
                marker_color="green",
                name="optimal MAE",
                showlegend=True,
                hovertemplate =None),
                row=1, col=3)
        fig.update_xaxes(title="NZV thresholds", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=3) 
        fig.update_yaxes(title="R-squared", showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', row=1, col=3)  

        fig.update_layout(height=500, width=1200, 
                        #paper_bgcolor="white",#'rgba(0,0,0,0)', 
                        plot_bgcolor='rgba(0,0,0,0)', 
                        margin=dict(l=0, r=0, b=0, pad=0),
                        title_text="Correlation",
                        hovermode="x")
        fig.show()

        correlation_threshold = correlation_thresholds[j]
    else:
        correlation_threshold = 1.0

    correlated_features = set()
    correlation_matrix = df[feature_cols].corr()
    for i in range(len(correlation_matrix.columns)):
        for j in range(i):
            if abs(correlation_matrix.iloc[i, j]) > correlation_threshold:
                colname = correlation_matrix.columns[i]
                correlated_features.add(colname)

    feature_cols_new = list(set(feature_cols)-correlated_features)
    
    print(f"Number of features left after removing features with correlation {correlation_threshold}: {len(feature_cols_new)}/{len(feature_cols)}")

    return feature_cols_new

def train(df, feature_cols, Regressor, parameters, param_grid, n_splits, file_name=None):
    """f"{PIPELINE_DIRECTORY}/model_NoTopImportant"""
    train1 = df[(df.healthy_reference==True)&(df.dataset_type=="Train")]

    X_train1, y_train1 = df2vectors(train1, feature_cols)

    rfr = Regressor(**parameters)
    gkf = list(GroupKFold(n_splits=n_splits).split(X_train1, y_train1, groups=train1["subjectID"].values))

    search = GridSearchCV(rfr, param_grid, cv=gkf)
    search.fit(X_train1, y_train1)

    estimator = search.best_estimator_
    if file_name:
        estimator.save_model(file_name)

    return estimator

def get_pvalue_regliner(df, group):
    _df = df.copy(deep=False)

    group_values = _df[group].unique()

    assert len(group_values) == 2, "the dataframe in statistical analysis needs to have only 2 unique groups to compare"

    df_stats = pd.DataFrame(data={"Input": list(_df.y.values),
                                  "Output": list(_df.y_pred.values),
                                  "Condition": list(_df[group].values)})

    return regliner(df_stats, {group_values[0]: 0, group_values[1]: 1})

def get_pvalue_permuspliner(df, group, degree=2):
    _df = df.copy(deep=False)

    group_values = _df[group].unique()

    assert len(group_values) == 2, "the dataframe in statistical analysis needs to have only 2 unique groups to compare"

    df_stats = pd.DataFrame(data={"Input": list(_df.y.values),
                                  "Output": list(_df.y_pred.values),
                                  "Condition":list(_df[group].values),
                                  "sampleID": list(_df["sampleID"].values)})

    result = permuspliner(df_stats, xvar="Input", yvar="Output", category="Condition", degree = degree, cases="sampleID", groups = group_values, perms = 500, test_direction = 'more', ints = 1000, quiet = True)

    return result["pval"]

def plot_trajectory(estimator, df, feature_cols, df_other=None, group=None, linear_difference=None, nonlinear_difference=None, 
                    plateau_area_start=None, limit_age=1200, start_age=0, 
                    time_unit_size=1, time_unit_name="days", img_file_name=None, 
                    degree=2, longitudinal_mode=None, longitudinal_showlegend=True,
                    patent=False, layout_settings=None, website=False, highlight_outliers=None, df_new=None,
                    plot_CI=False, plot_PI=True, dtick=2, stats_table=True, PI_percentage=95):
    """Trajectory line with performance stats and many other settings

    estimator: sklearn models, CatBoostRegressor, etc.
        Model for the trajectory line.
    df: pd.DataFrame
        Reference dataset, validation data.
    feature_cols: list
        Feature columns needed for the estimator of the trajectory.
    df_other: pd.DataFrame
        Other dataset (not the reference)
    degree: int
        The degree for the polyfit line (trajectory).
    traj_color: str
        Color for the trajectory line. Usually green, but if it is a plot for IP, we need gray.
    group: bool
        To plot the colors based on the group.
    linear_difference: bool
        If True, plot the linear lines for each of the countries and calculate the significant difference between
        these linear lines.
    nonlinear_difference: bool
        If True, plot the spline lines for each of the countries and calculate the significant difference between
        these splines.
    plateau_area_start: int
        The day when the plateau starts (shade that area gray then). Usuallyafter 2-3 year, i.e. 720+ days.
    limit_age: int
        Days when to end the trajectory (used to cut the first few days and last days for infants)
    start_age: int
        Days when to start the trajectory (used to cut the first few days and last days for infants)
    time_unit_name: str
        Name of the time unit (e.g. month, year, etc.)
    time_unit_size: int
        Number of days in a new time definition (e.g. if we want to deal with months, then time_unit_in_days=30, for the year, time_unit_in_days=365)
    img_file_name: str
        Name of the file where to save the plot of trajectory.
    nboot: int
        Number of bootstrap samples from the data.

    Returns
    -------
    fig: plotly object
        To continue furhter plots in case needed.
    outliers: list
        List of outliers.
    mae: float
        The MAE error.
    r2: float
        The R^2 metric.
    pi_median: float
        Prediction interval median value across the ages.
    """
    df = df.sort_values(by="age_at_collection")
    
    fig = go.Figure()

    limit_age_max = int(max(df["age_at_collection"]))+1

    layout_settings_default = dict(
        height=900, 
        width=1000,
        barmode='stack', 
        uniformtext=dict(mode="hide", minsize=10),
        plot_bgcolor='rgba(0,0,0,0)',
        paper_bgcolor='rgba(0,0,0,0)',  
        margin=dict(l=0, r=0, b=0, pad=0),
        title_text="Microbiome Trajectory"
    )

    if patent:
        traj_color = "0,0,0"
        colors = px.colors.sequential.Greys
        outlier_color = '0,0,0'
        outlier_size = 15
        marker_outlier=dict(size=25, color=f'rgba({outlier_color},0.95)', symbol="star-open", line_width=4)
        layout_settings_default["height"]=900
        layout_settings_default["width"]=1100
        layout_settings_default["font"] = dict(
                #family="Courier New, monospace",
                size=20,
                #color="RebeccaPurple"
            )
    else:
        traj_color = "26,150,65"
        colors = px.colors.qualitative.Plotly
        outlier_color = '255,0,0'
        outlier_size = 15  
        marker_outlier=dict(size=20, color=f'rgba({outlier_color},0.95)', symbol="star-open", line_width=4)  
        
    if layout_settings is None:
        layout_settings = {}
    layout_settings_final = {**layout_settings_default, **layout_settings}
    
    if plateau_area_start is not None:
        # shaded area where plateau is expected
        if plateau_area_start/time_unit_size < limit_age/time_unit_size:
            _x = np.linspace(plateau_area_start, limit_age_max, 10)/time_unit_size
            fig.add_trace(go.Scatter(
                x=list(_x)+list(_x[::-1]),
                y=list(np.zeros(10))+list(np.ones(10)*limit_age/time_unit_size+1),
                fill='toself',
                fillcolor='rgba(220,220,220,0.5)',
                line_color='rgba(220,220,220,0.5)',
                showlegend=True,
                name=f"sample at time > {plateau_area_start} days",
            ))
    
    if group is not None:
        longitudinal_showlegend = False

    fig, ret_val, mae, r2, pi_median, _, _, _ = plot_1_trajectory(fig, estimator, df, feature_cols, limit_age, time_unit_size, time_unit_name, traj_color=traj_color, traj_label="reference", 
                                                                            plateau_area_start=plateau_area_start, limit_age_max=limit_age_max, longitudinal_mode=longitudinal_mode, 
                                                                            longitudinal_showlegend=longitudinal_showlegend, highlight_outliers=highlight_outliers, marker_outlier=marker_outlier, df_new=df_new,
                                                                            plot_CI=plot_CI, plot_PI=plot_PI, PI_percentage=PI_percentage)
    
    X, y = df2vectors(df, feature_cols)
    y_pred = estimator.predict(X)

    if PI_percentage == 95:
        percent = 0.975   
    elif  PI_percentage == 90:
        percent = 0.95
    elif PI_percentage == 80:
        percent = 0.9
    else:
        raise NotImplemented("The percentage not implemented!")
    
    y = np.array(y)/time_unit_size
    y_pred = np.array(y_pred)/time_unit_size
    
    df["y"] = y
    df["y_pred"] = y_pred
    
    equation = lambda a, b: np.polyval(a, b) 
        
    # Data
    if group is not None:
        colors_rgb = [tuple(int(h.lstrip('#')[i:i+2], 16) for i in (0, 2, 4)) for h in colors]
        
        # linear difference between the groups
        if linear_difference:
            
            for i, group_trace in enumerate(df[group].unique()):
                idx = np.where(df[group]==group_trace)[0]
                _df = df[df[group]==group_trace]
                #xdata = np.linspace(0, limit_age_max//time_unit_size+1, limit_age_max//time_unit_size+1)
                xdata = np.linspace(y[idx].min()/time_unit_size-dtick, y[idx].max()/time_unit_size+dtick, limit_age_max+1)

                # lines
                _p, _cov = np.polyfit(y[idx], y_pred[idx], 1, cov=True)    
                fig.add_trace(go.Scatter(
                    x=xdata,
                    y=equation(_p, xdata),
                    mode="lines",
                    line = dict(width=3, dash='dash', color=colors[i]),
                    marker=dict(size=10, color=colors[i]),
                    showlegend=True,
                    legendgroup=group_trace,
                    name=f"{group}={group_trace}",
                    text=list(_df["sampleID"].values), 
                    hovertemplate = f'<b>Group ({group}): {group_trace}</b><br>',
                    hoveron="points"
                ))
                # points    
                fig.add_trace(go.Scatter(
                    x=y[idx],
                    y= y_pred[idx],
                    mode="markers",
                    line = dict(width=3, dash='dash', color=colors[i]),
                    marker=dict(size=10, color=colors[i]),
                    showlegend=True,
                    legendgroup=group_trace,
                    name=f"sample with {group}={group_trace}",
                    text=list(_df["sampleID"].values), 
                    hovertemplate = '<b>Healthy reference sample</b><br><br>'+
                                    f'<b>Group ({group}): {group_trace}</b><br>'+
                                    '<b>SampleID</b>: %{text}<br>'+
                                    '<b>Age</b>: %{x:.2f}'+
                                    '<br><b>MMI</b>: %{y}<br>',
                    hoveron="points"
                ))
            
            group_vals = df[group].unique()
            comb = combinations(group_vals, 2)
            ret_val += "<b>Linear p-value (k, n)</b>:"
            for c in list(comb):
                _df = df[(df[group].isin(c))&(df.age_at_collection<limit_age)]
                pval_k, pval_n = get_pvalue_regliner(_df, group)
                ret_val += f"<br>{group} {c[0]} vs. {c[1]}: ({pval_k:.3f}, {pval_n:.3f})"
        
        # non-linear difference between the groups (splines)
        elif nonlinear_difference:
            for i, group_trace in enumerate(df[group].unique()):
                idx = np.where(df[group]==group_trace)[0]
                _df = df[df[group]==group_trace]
                
                xdata = y[idx]
                ydata = y_pred[idx]
                _p, _cov = np.polyfit(xdata, ydata, degree, cov=True)    
                _y_model = equation(_p, ydata)                             # model using the fit parameters; NOTE: parameters here are coefficients

                # Statistics
                n = ydata.size                                             # number of observations
                m = _p.size                                                # number of parameters
                dof = n - m                                                # degrees of freedom
                t = stats.t.ppf(percent, n - m)                              # used for CI and PI bands

                # Estimates of Error in Data/Model
                resid = ydata - _y_model                           
                chi2 = np.sum((resid / _y_model)**2)                       # chi-squared; estimates error in data
                chi2_red = chi2 / dof                                      # reduced chi-squared; measures goodness of fit
                s_err = np.sqrt(np.sum(resid**2) / dof)                    # standard deviation of the error

                #x2 = np.linspace(0, limit_age_max/time_unit_size, limit_age_max+1) #np.linspace(np.min(x), np.max(x), 100)
                x2 = np.linspace(xdata.min()/time_unit_size-dtick, xdata.max()/time_unit_size+dtick, limit_age_max+1)
                y2 = equation(_p, x2)
                pi = t * s_err * np.sqrt(1 + 1/n + (x2 - np.mean(y))**2 / np.sum((y - np.mean(y))**2))   
                
                # mean prediction
                fig.add_trace(go.Scatter(
                    x=x2, y=y2,
                    line_color=colors[i],
                    name=f"trajectory for {group}={group_trace}",
                    legendgroup=group_trace,
                ))
                # prediction interval
                fig.add_trace(go.Scatter(
                    x=list(x2)+list(x2[::-1]),
                    y=list(y2-pi)+list(y2+pi)[::-1],
                    fill='toself',
                    fillcolor=f'rgba{tuple(list(colors_rgb[i])+[0.15])}',
                    line_color=f'rgba{tuple(list(colors_rgb[i])+[0.25])}',
                    legendgroup=group_trace, 
                    showlegend=False,
                    name="95% Prediction Interval"
                ))
                # points    
                fig.add_trace(go.Scatter(
                    x=y[idx],
                    y= y_pred[idx],
                    mode="markers",
                    line = dict(width=3, dash='dash', color=colors[i]),
                    marker=dict(size=10, color=colors[i]),
                    showlegend=True,
                    legendgroup=group_trace,
                    name=f"sample with {group}={group_trace}",
                    text=list(_df["sampleID"].values), 
                    hovertemplate = '<b>Healthy reference sample</b><br><br>'+
                                    f'<b>Group ({group}): {group_trace}</b><br>'+
                                    '<b>SampleID</b>: %{text}<br>'+
                                    '<b>Age</b>: %{x:.2f}'+
                                    '<br><b>MMI</b>: %{y}<br>',
                    hoveron="points"
                )) 
            
            group_vals = df[group].unique()
            comb = combinations(group_vals, 2)
            ret_val += "<b>Nonlinear p-value</b>:"
            for c in list(comb):
                _df = df[(df[group].isin(c))]  #&(df.age_at_collection<limit_age)&(df.age_at_collection>start_age)
                error_cnt = 3
                while error_cnt > 0:
                    try:
                        pval = get_pvalue_permuspliner(_df, group, degree=degree)
                        error_cnt = -1
                    except:
                        error_cnt -= 1
                if error_cnt == -1:
                    ret_val += f"<br>{group} {c[0]} vs. {c[1]}: {pval:.3f}"
                else:
                    ret_val += f"<br>{group} {c[0]} vs. {c[1]}: not available"

                    
        else:
            # longitudinal, but color based on the group
            for i, group_trace in enumerate(df[group].unique()):
                idx = np.where(df[group]==group_trace)[0]
                _df = df[df[group]==group_trace]
                color = colors[i]
                for j, trace in enumerate(_df["subjectID"].unique()):
                    idx2 = np.where(_df["subjectID"]==trace)[0]
                    fig.add_trace(go.Scatter(
                        x=y[idx][idx2],
                        y=y_pred[idx][idx2],
                        mode="lines+markers",
                        line = dict(width=3, dash='dash', color=color),
                        marker=dict(size=10, color=color),
                        showlegend=True if j == 0 else False,
                        legendgroup=group_trace,
                        name=group_trace,
                        text=list(_df["sampleID"].values[idx2]), 
                        hovertemplate = '<b>Healthy reference sample</b><br><br>'+
                                        f'<b>Group ({group}): {group_trace}</b><br>'+
                                        '<b>SampleID</b>: %{text}<br>'+
                                        '<b>Age</b>: %{x:.2f}'+
                                        '<br><b>MMI</b>: %{y}<br>',
                        hoveron="points"
                    ))
      
    ###########################################################################
    X_other, y_other, y_other_pred = None, None, None
    if df_other is not None:
        
        df_other = df_other.sort_values(by="age_at_collection")

        X_other, y_other = df2vectors(df_other, feature_cols)
        y_other_pred = estimator.predict(X_other)
        
        y_other = np.array(y_other)/time_unit_size
        y_other_pred = np.array(y_other_pred)/time_unit_size
        
        for trace in df_other["subjectID"].unique():
            idx = np.where(df_other["subjectID"].values==trace)[0]

            fig.add_trace(go.Scatter(
                x=y_other[idx],
                y=y_other_pred[idx],
                mode=longitudinal_mode,
                line=dict(width=3, dash='dash'),
                marker=dict(size=outlier_size, color=f'rgba({outlier_color},0.95)'),
                showlegend=longitudinal_showlegend,
                name=trace,
                text=list(df_other["sampleID"].values[idx]), 
                hovertemplate = '<b>Other sample</b><br><br>'+
                                '<b>SampleID</b>: %{text}<br>'+
                                '<b>Age</b>: %{x:.2f}<br>'+
                                '<b>MMI</b>: %{y}<br>',
                hoveron="points"
            ))


    fig.update_xaxes(title=f"Age [{time_unit_name}]", range=(start_age//time_unit_size, limit_age//time_unit_size), 
                    tick0=start_age//time_unit_size, dtick=dtick, 
                    showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray') 
    fig.update_yaxes(title=f"Microbiome Maturation Index [{time_unit_name}]", range=(start_age//time_unit_size, limit_age//time_unit_size), 
                    tick0=start_age//time_unit_size, dtick=dtick, 
                    showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray')  
    
    fig.update_layout(**layout_settings_final)
    
    if not patent:
        if stats_table:
            fig.update_layout(go.Layout(
                annotations=[
                    go.layout.Annotation(
                        text=ret_val,
                        align='left',
                        showarrow=False,
                        xref='paper',
                        yref='paper',
                        x=.99,
                        y=.01,
                        bordercolor='black',
                        bgcolor='white',
                        borderwidth=0.5,
                        borderpad=8
                    )
                ]
            ))
    else:
        print(ret_val)
    
    if img_file_name:
        pathlib.Path('/'.join(img_file_name.split('/')[:-1])).mkdir(parents=True, exist_ok=True)
    
        if "html" in img_file_name.lower(): 
            fig.write_html(img_file_name)
        elif any(extension in img_file_name for extension in ["png", "jpeg", "webp", "svg", "pdf", "eps" ]):
            fig.write_image(img_file_name)
        else:
            raise Exception(f"Extension {img_file_name.split('.')[-1]} is not implemented")
    

    
    if not website:
        fig.show()
    
    return fig, mae, r2, pi_median




def plot_1_trajectory(fig, estimator, df, bacteria_names, limit_age, time_unit_size, time_unit_name, traj_label, plateau_area_start, traj_color, limit_age_max, df_new=None, degree=2, longitudinal_mode=None, longitudinal_showlegend=True, 
                    fillcolor_alpha=0.3, highlight_outliers=None, marker_outlier=None, plot_CI=False, plot_PI=True, PI_percentage=95): 
    """
    longitudinal_mode: str
        How a longitudinal data is plotted: markers+lines, lines, markers, etc.
    Reference:
    - https://nbviewer.jupyter.org/github/demotu/BMC/blob/master/notebooks/CurveFitting.ipynb
    """
    if df_new is None:
        df_new = df.copy()
    
    df = df.sort_values(by="age_at_collection")
    X, y = df2vectors(df, bacteria_names)
    y_pred = estimator.predict(X)
    y = np.array(y)/time_unit_size
    y_pred = np.array(y_pred)/time_unit_size
    
    df_new = df_new.sort_values(by="age_at_collection")
    X_new, y_new = df2vectors(df_new, bacteria_names)
    y_pred_new = estimator.predict(X_new)
    y_new = np.array(y_new)/time_unit_size
    y_pred_new = np.array(y_pred_new)/time_unit_size

    mae   = round(np.mean(abs(y_pred - y)), 2)
    r2    = r2_score(y, y_pred)
    coeff = stats.pearsonr(y_pred, y)
    
    # performace calculated until limit_age
    idx       = np.where(y < limit_age/time_unit_size)[0]
    mae_idx   = round(np.mean(abs(y_pred[idx] - y[idx])), 2)
    r2_idx    = r2_score(y[idx], y_pred[idx])
    coeff_idx = stats.pearsonr(y_pred[idx], y[idx])
    ret_val = "<b>Performance Information</b><br>"
    ret_val += f'MAE: {mae_idx}<br>'
    ret_val += f'R^2: {r2_idx:.3f}<br>'
    ret_val += f"Pearson: {coeff_idx[0]:.3f}, 2-tailed p-value: {coeff_idx[1]:.2e}<br>"
    
    if plateau_area_start is not None:
        idx       = np.where(y < plateau_area_start/time_unit_size)[0]
        mae_idx   = round(np.mean(abs(y_pred[idx] - y[idx])), 2)
        r2_idx    = r2_score(y[idx], y_pred[idx])
        coeff_idx = stats.pearsonr(y_pred[idx], y[idx])
    
        ret_val += f"<b>Performance Information < {plateau_area_start} days</b><br>"
        ret_val += f'MAE: {mae_idx}<br>'
        ret_val += f'R^2: {r2_idx:.3f}<br>'
        ret_val += f"Pearson: {coeff_idx[0]:.3f}, 2-tailed p-value: {coeff_idx[1]:.2e}<br>"

    if PI_percentage == 95:
        percent = 0.975   
    elif  PI_percentage == 90:
        percent = 0.95
    elif PI_percentage == 80:
        percent = 0.9
    else:
        raise NotImplemented("The percentage not implemented!")
        
    # Plot data
    equation = lambda a, b: np.polyval(a, b) 

    # Stats
    p, cov = np.polyfit(y, y_pred, degree, cov=True)           # parameters and covariance from of the fit of 1-D polynom.
    y_model = equation(p, y)                                   # model using the fit parameters; NOTE: parameters here are coefficients

    # Statistics
    n = y_pred.size                                            # number of observations
    m = p.size                                                 # number of parameters
    dof = n - m                                                # degrees of freedom
    t = stats.t.ppf(percent, n - m)                              # used for CI and PI bands

    # Estimates of Error in Data/Model
    resid = y_pred - y_model                           
    chi2 = np.sum((resid / y_model)**2)                        # chi-squared; estimates error in data
    chi2_red = chi2 / dof                                      # reduced chi-squared; measures goodness of fit
    s_err = np.sqrt(np.sum(resid**2) / dof)                    # standard deviation of the error

    # Confidence Interval (select one)
    if plot_CI:
        nboot = 500
        bootindex = sp.random.randint
        
        for b in range(nboot):
            resamp_resid = resid[bootindex(0, len(resid) - 1, len(resid))]
            # Make coeffs of for polys
            pc = np.polyfit(y, y_pred + resamp_resid, degree)                   
            # Plot bootstrap cluster
            idx = np.argsort(y)
            _x = y[idx]
            _y = np.polyval(pc, _x)
            
            if b < nboot-1:
                fig.add_trace(go.Scatter(
                    x=_x,
                    y=_y,
                    mode="lines",
                    line = dict(width=5),
                    fillcolor=f'rgba({traj_color},{3.0 / float(nboot):.2f})',
                    line_color=f'rgba({traj_color},{3.0 / float(nboot):.2f})',
                    showlegend=False,
                    name="Confidence Interval",
                ))
            else:
                fig.add_trace(go.Scatter(
                    x=_x,
                    y=_y,
                    mode="lines",
                    line = dict(width=5),
                    fillcolor=f'rgba({traj_color},{0.1:.2f})',
                    line_color=f'rgba({traj_color},{0.1:.2f})',
                    showlegend=True,
                    name=f"{PI_percentage}% Confidence Interval",
                    legendgroup="trajectory"
                ))

    x2 = np.linspace(0, limit_age_max/time_unit_size, limit_age_max+1) #np.linspace(np.min(x), np.max(x), 100)
    y2 = equation(p, x2)

    # Prediction Interval
    pi = t * s_err * np.sqrt(1 + 1/n + (x2 - np.mean(y))**2 / np.sum((y - np.mean(y))**2))   
    pi_mean       = np.mean(pi)
    pi_median     = np.median(pi)
    
    ret_val += "<b>Statistics</b><br>"
    ret_val += f"Chi^2: {chi2:.2f}<br>"
    ret_val += f"Reduced chi^2: {chi2_red:.2f}<br>"
    ret_val += f"Standard deviation of the error: {s_err:.2f}<br>"
    ret_val += f"Prediction interval:<br> mean={pi_mean:.2f}, median={pi_median:.2f}<br>"
    
    if plateau_area_start is not None:
        idx           = np.where(x2 < plateau_area_start)[0]
        pi_mean_idx   = np.mean(pi[idx])
        pi_median_idx = np.median(pi[idx])
        ret_val += f"Prediction interval < {plateau_area_start} {time_unit_name}:<br> mean={pi_mean_idx:.2f}, median={pi_median_idx:.2f}<br>"

    # mean prediction
    fig.add_trace(go.Scatter(
        x=x2, y=y2,
        line_color=f'rgba({traj_color},1.)',
        name=f"{traj_label.title()} trajectory",
        legendgroup="trajectory"
    ))
    # prediction interval
    if plot_PI:
        fig.add_trace(go.Scatter(
            x=list(x2)+list(x2[::-1]),
            y=list(y2-pi)+list(y2+pi)[::-1],
            fill='toself',
            fillcolor=f'rgba({traj_color},{fillcolor_alpha})',
            line_color=f'rgba({traj_color},{fillcolor_alpha+0.2})',
            showlegend=True,
            name=f"{PI_percentage}% Prediction Interval",
            legendgroup="trajectory"
        ))
        

    if highlight_outliers is not None:
        idx = np.where(df_new["sampleID"].astype(str).isin(highlight_outliers))[0]
        fig.add_trace(go.Scatter(
            x=y_new[idx],
            y=y_pred_new[idx],
            mode="markers",
            marker=marker_outlier,
            showlegend=True,
            name="Outliers",
            text=list(df_new["sampleID"].values[idx]), 
            hovertemplate = '<b>Healthy reference sample outside the healthy region</b><br><br>'+
                            '<b>SampleID</b>: %{text}<br>'+
                            '<b>Age</b>: %{x:.2f}'+
                            '<br><b>MMI</b>: %{y}<br>',
            hoveron="points"
        ))
    
    if longitudinal_mode is not None:
        if longitudinal_mode == "markers":
            fig.add_trace(go.Scatter(
                    x=y_new,
                    y=y_pred_new,
                    mode=longitudinal_mode,
                    line=dict(width=3, dash='dash', color=f'rgba({traj_color},0.65)'),
                    marker=dict(size=10, color=f'rgba({traj_color},0.65)'),
                    showlegend=True,
                    name=f"{traj_label.title()} samples",
                    text=list(df_new["sampleID"].values), 
                    hovertemplate = f'<b>{traj_label.title()} reference sample</b><br><br>'+
                                    '<b>SampleID</b>: %{text}<br>'+
                                    '<b>Age</b>: %{x:.2f}'+
                                    '<br><b>MMI</b>: %{y}<br>',
                    hoveron="points"
                ))
        else:
            # longitudinal - line per subject
            for trace in df_new["subjectID"].unique():
                idx = np.where(df_new["subjectID"]==trace)[0]
                fig.add_trace(go.Scatter(
                    x=y_new[idx],
                    y=y_pred_new[idx],
                    mode=longitudinal_mode,
                    line=dict(width=3, dash='dash', color=f'rgba({traj_color},0.65)'),
                    marker=dict(size=10, color=f'rgba({traj_color},0.65)'),
                    showlegend=longitudinal_showlegend,
                    name=trace,
                    text=list(df_new["sampleID"].values[idx]), 
                    hovertemplate = f'<b>{traj_label.title()} reference sample</b><br><br>'+
                                    '<b>SampleID</b>: %{text}<br>'+
                                    '<b>Age</b>: %{x:.2f}'+
                                    '<br><b>MMI</b>: %{y}<br>',
                    hoveron="points"
                ))
    
    
    return fig, ret_val, mae, r2, pi_median, x2, pi, y2

def plot_2_trajectories(estimator_ref, val1, val2, feature_cols, degree=2, plateau_area_start=2, limit_age=1200, start_age=0, time_unit_size=1, time_unit_name="days", 
                        linear_pval=False, nonlinear_pval=False, img_file_name=None, longitudinal_mode="markers+lines", 
                        website=False, plot_CI=False, plot_PI=True, layout_settings=None, stats_table=True, dtick=2,PI_percentage=95):
    val1 = val1.sort_values(by="age_at_collection")
    X1, y1 = df2vectors(val1, feature_cols)
    y_pred1 = estimator_ref.predict(X1)
    sid1 = val1["sampleID"].values
    
    val2 = val2.sort_values(by="age_at_collection")
    X2, y2 = df2vectors(val2, feature_cols)
    y_pred2 = estimator_ref.predict(X2)
    sid2 = val2["sampleID"].values
    

    fig = go.Figure()
    ret_val = ""
    
    limit_age_max1 = int(max(val1["age_at_collection"]))+1
    limit_age_max2 = int(max(val2["age_at_collection"]))+1
    limit_age_max = max(limit_age_max1, limit_age_max2)

    if plateau_area_start is not None:
        # shaded area where plateau is expected
        if plateau_area_start//time_unit_size+1 < limit_age//time_unit_size+1:
            _x = np.linspace(plateau_area_start, limit_age_max, 10)//time_unit_size+1
            fig.add_trace(go.Scatter(
                x=list(_x)+list(_x[::-1]),
                y=list(np.zeros(10))+list(np.ones(10)*limit_age//time_unit_size+1),
                fill='toself',
                fillcolor='rgba(220,220,220,0.5)',
                line_color='rgba(220,220,220,0.5)',
                showlegend=True,
                name=f"sample at time > {plateau_area_start} days",
            ))
    else:
        plateau_area_start = limit_age
    
    
    fig, ret_val1, _, _, _, _, _, _ = plot_1_trajectory(fig, estimator_ref, val1, feature_cols, limit_age, time_unit_size, time_unit_name, traj_color="0,0,255", traj_label="reference", 
                                                            plateau_area_start=plateau_area_start, limit_age_max=limit_age_max, longitudinal_mode=longitudinal_mode, longitudinal_showlegend=False,
                                                            plot_CI=plot_CI, plot_PI=plot_PI, PI_percentage=PI_percentage)
    fig, ret_val2, _, _, _, _, _, _ = plot_1_trajectory(fig, estimator_ref, val2, feature_cols, limit_age, time_unit_size, time_unit_name, traj_color="255,0,0", traj_label="other", 
                                                            plateau_area_start=plateau_area_start, limit_age_max=limit_age_max, longitudinal_mode=longitudinal_mode, longitudinal_showlegend=False,
                                                            plot_CI=plot_CI, plot_PI=plot_PI, PI_percentage=PI_percentage)

              
    # dataframe will be used for linear and nonlinear p-value calculation
    df = pd.DataFrame(data={"y":np.concatenate([y1, y2]), "y_pred":np.concatenate([y_pred1, y_pred2]),
                            "diftimeunit_y":np.concatenate([y1, y2]), "diftimeunit_y_pred":np.concatenate([y_pred1, y_pred2]),
                            "sampleID":list(sid1)+list(sid2),
                            "label": ["reference"]*len(y1) + ["other"]*len(y2)})

    if linear_pval:
        df = df[(df.y<plateau_area_start) & (start_age<df.y)]
        
        equation = lambda a, b: np.polyval(a, b) 
        xx = np.linspace(0, plateau_area_start, 10)/time_unit_size+1

        
        # lines
        p1, _cov = np.polyfit(df[df.label=="reference"].y.values//time_unit_size+1, df[df.label=="reference"].y_pred.values//time_unit_size+1, 1, cov=True)  
        fig.add_trace(go.Scatter(
            x=xx,
            y=equation(p1, xx),
            mode="lines",
            line = dict(width=5, dash='dash', color="rgba(0,0,255,1.0)"),
            marker=dict(size=10, color='red'),
            showlegend=True,
            name='Reference linear line',
        ))
        
        p2, _cov = np.polyfit(df[df.label=="other"].y.values//time_unit_size+1, df[df.label=="other"].y_pred.values//time_unit_size+1, 1, cov=True)         
        fig.add_trace(go.Scatter(
            x=xx,
            y=equation(p2, xx),
            mode="lines",
            line = dict(width=5, dash='dash', color='red'),
            marker=dict(size=10, color="rgba(255,0,0,1.0)"),
            showlegend=True,
            name='Other linear line',
        ))
        pval_k, pval_n = get_pvalue_regliner(df, group="label")
        ret_val += f"<b>Linear lines difference:</b><br>p = {pval_k:.3f}, {pval_n:.3f}"

    if nonlinear_pval:
        df = df[(df.y<limit_age) & (start_age<df.y)]
        
        pval = get_pvalue_permuspliner(df, group="label", degree=degree)
        
        ret_val += f"<b>Splines difference:</b><br>p = {pval:.3f}"
        
    fig.update_xaxes(title=f"Age [{time_unit_name}]", range=(start_age/time_unit_size, limit_age/time_unit_size), 
                     tick0=start_age/time_unit_size, dtick=dtick, showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray') 
    fig.update_yaxes(title=f"Microbiome Maturation Index [{time_unit_name}]", range=(start_age/time_unit_size, limit_age/time_unit_size), 
                     tick0=start_age/time_unit_size, dtick=dtick, showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray')  
    

    layout_settings_default = dict(
        height=900, 
        width=1000,
        plot_bgcolor='rgba(0,0,0,0)', 
        paper_bgcolor='rgba(0,0,0,0)', 
        margin=dict(l=0, r=0, b=0, pad=0),
        title_text="Microbiome Trajectory"
    )

    if layout_settings is None:
        layout_settings = {}
    layout_settings_final = {**layout_settings_default, **layout_settings}
    
    fig.update_layout(**layout_settings_final)
    
    if stats_table:
        fig.update_layout(go.Layout(
            annotations=[
                go.layout.Annotation(
                    text=ret_val1,
                    font = dict(size = 10),
                    align='left',
                    showarrow=False,
                    xref='paper',
                    yref='paper',
                    x=0.5,
                    y=.01,
                    bordercolor='black',
                    bgcolor='rgba(0,0,220,0.5)',
                    borderwidth=0.5,
                    borderpad=8
                ),
                go.layout.Annotation(
                    text=ret_val2,
                    font = dict(size = 10),
                    align='left',
                    showarrow=False,
                    xref='paper',
                    yref='paper',
                    x=1.0,
                    y=.01,
                    bordercolor='black',
                    bgcolor='rgba(220,0,0,0.5)',
                    borderwidth=0.5,
                    borderpad=8
                ),
                go.layout.Annotation(
                    text=ret_val,
                    font = dict(size = 15),
                    align='left',
                    showarrow=False,
                    xref='paper',
                    yref='paper',
                    x=.01,
                    y=.99,
                    bordercolor='black',
                    bgcolor='white',
                    borderwidth=0.5,
                    borderpad=8
                )
            ]
        ), xaxis=dict(domain=[0.1, 0.1]))
    
    if img_file_name:
        fig.write_html(img_file_name)

    if not website:
        fig.show()
    
    return fig





from sklearn.preprocessing import StandardScaler
from sklearn.ensemble import IsolationForest
from sklearn.preprocessing import StandardScaler
from sklearn.ensemble import IsolationForest
from sklearn import svm
from microbiome.helpers import df2vectors
import pandas as pd
import numpy as np
import scipy.stats as stats


def pi_anomaly_detection(estimator, df_all, feature_columns, degree, df_new=None, PI_percentage=95):
    """
    Prediction Interval anomaly detection. We consider samples outside prediction interval to be outliers.
    
    Parameters
    ----------
    estimator: sklearn mordl, etc.
        Estimator used for microbiome trajectory.
    df: pd.DataFrame
        Dataframe containing data
    feature_columns: float
        List of taxa on which the model/estimator was trained on (usualy)
    degree: int
        Degree of the microbiome trajectory poly line.
    
    Returns
    -------
    outliers :list
        List of sampleIDs that are anomalous.
    """
    df = df_all.copy()

    if PI_percentage == 95:
        percent = 0.975   
    elif  PI_percentage == 90:
        percent = 0.95
    elif PI_percentage == 80:
        percent = 0.9
    else:
        raise NotImplemented("The percentage not implemented!")

    def _inner_pi_anomaly_detection(df_model, df_model_new, degree):
        """ Brain of this anomaly detection
        Parameters
        ----------
        df_model: pd.DataFrame
            Dataframe containing x-axis (age_at_collection, which we call y), y-axis (MMI, which we call y_pred), and unique sample IDs (sampleID)
        degree: float
            Polynomial degree of this microbiome trajectory
            
        Returns
        -------
        outliers :list
            List of sampleIDs that are anomalous.
        """
        limit_age_max = int(max(df_model["y"]))+1

        equation = lambda a, b: np.polyval(a, b) 

        # Stats
        p, cov = np.polyfit(df_model["y"].values, df_model["y_pred"].values, degree, cov=True)           # parameters and covariance from of the fit of 1-D polynom.
        y_model = equation(p, df_model["y"].values)                                   # model using the fit parameters; NOTE: parameters here are coefficients

        # Statistics
        n = df_model["y_pred"].values.size                                            # number of observations
        m = p.size                                                 # number of parameters
        dof = n - m                                                # degrees of freedom
        t = stats.t.ppf(percent, n - m)                              # used for CI and PI bands

        # Estimates of Error in Data/Model
        resid = df_model["y_pred"].values - y_model                           
        chi2 = np.sum((resid / y_model)**2)                        # chi-squared; estimates error in data
        chi2_red = chi2 / dof                                      # reduced chi-squared; measures goodness of fit
        s_err = np.sqrt(np.sum(resid**2) / dof)                    # standard deviation of the error
    
        x2 = np.linspace(0, limit_age_max, limit_age_max+1) #np.linspace(np.min(x), np.max(x), 100)
        y2 = equation(p, x2)

        # Prediction Interval
        pi = t * s_err * np.sqrt(1 + 1/n + (x2 - np.mean(df_model["y"].values))**2 / np.sum((df_model["y"].values - np.mean(df_model["y"].values))**2))   
        pi_mean       = np.mean(pi)
        pi_median     = np.median(pi)
        
        # highlight_outliers - list of outliers to highlight, or True or False to plot all
        p_lower, _ = np.polyfit(x2, y2-pi, degree, cov=True)
        p_upper, _ = np.polyfit(x2, y2+pi, degree, cov=True)  
        outliers = set()
        for i in range(len(df_model_new["y"].values)):
            if df_model["y_pred"].values[i] > equation(p_upper, df_model_new["y"].values[i]) or df_model["y_pred"].values[i] < equation(p_lower, df_model_new["y"].values[i]):
                outliers.add(str(df_model_new.iloc[i]["sampleID"]))

        
        return list(outliers)
    
    df = df.sort_values(by="age_at_collection")
    X, y = df2vectors(df, feature_columns)
    y_pred = estimator.predict(X)
    
    if df_new is None:
        df_new = df.copy()
        
    df_new = df_new.sort_values(by="age_at_collection")
    X_new, y_new = df2vectors(df_new, feature_columns)
    y_pred_new = estimator.predict(X_new)
    
    df_model = pd.DataFrame(data={"y":y, "y_pred":y_pred, 'sampleID': df.sampleID})
    df_model_new = pd.DataFrame(data={"y":y_new, "y_pred":y_pred_new, 'sampleID': df_new.sampleID})
    
    outliers = _inner_pi_anomaly_detection(df_model, df_model_new, degree)
        
    return outliers    



def lpf_anomaly_detection(estimator, df_all, feature_columns, num_of_stds, window):
    """
    Low-pass filter (LPF) to detect anomalies in a time series. We unite all longitudinal data into one signal. 
    If we had more smaples per subjects, we could do this lpf anomaly detection for every subject separately.
    Parameters
    ----------
    df: pd.DataFrame
        Dataframe containing data
    num_of_stds: float
        Number of standard deviations of z-score that will be considered anomalous if higher.
    window: int
        Number of observations used for calculating the statistic.
    
    Returns
    -------
    outliers :list
        List of sampleIDs that are anomalous.
    """
    df = df_all.copy()
    def _inner_lpf_anomaly_detection(df_model, num_of_stds, window):
        """ Brain of this anomaly detection
        Parameters
        ----------
        df: pd.DataFrame
            Dataframe containing x-axis (age_at_collection, which we call y), y-axis (MMI, which we call y_pred), and unique sample IDs (sampleID)
        num_of_stds: float
            Number of standard deviations of z-score that will be considered anomalous if higher.
        window: int
            Number of observations used for calculating the statistic.
            
        Returns
        -------
        outliers :list
            List of sampleIDs that are anomalous.
        """
        df_model = df_model.sort_values(by="y")
        df_model["y_pred_rolling_avg"]=df_model["y_pred"].rolling(window=window, min_periods=1, center=True).mean()
        df_model["y_pred_rolling_std"]=df_model["y_pred"].rolling(window=window, min_periods=1, center=True).std(skipna=True)
        df_model["lpf_anomaly"]= abs(df_model["y_pred"]-df_model["y_pred_rolling_avg"])>(num_of_stds*df_model["y_pred_rolling_std"])
        return df_model[df_model["lpf_anomaly"]==True].sampleID.values.tolist()
    
    X, y = df2vectors(df, feature_columns)
    y_pred = estimator.predict(X)
    df_model = pd.DataFrame(data={"y":y, "y_pred":y_pred, 'sampleID': df.sampleID})
    
    outliers = _inner_lpf_anomaly_detection(df_model, num_of_stds, window)
        
    return outliers    


def if_anomaly_detection(estimator, df_all, feature_columns, outliers_fraction, window, anomaly_columns):
    """
    Isolation Forest (IF) to detect anomalies in a time series. We unite all longitudinal data into one signal. 
    If we had more smaples per subjects, we could do this lpf anomaly detection for every subject separately.
    Parameters
    ----------
    df: pd.DataFrame
        Dataframe containing data
    num_of_stds: float
        Number of standard deviations of z-score that will be considered anomalous if higher.
    window: int
        Number of observations used for calculating the statistic.
    anomaly_column: list
        Two options: `y_pred`, which ignores the timebox, and `y_pred_zcore`, or both.
    
    Returns
    -------
    outliers :list
        List of sampleIDs that are anomalous.
    """
    df = df_all.copy()
    def _inner_if_anomaly_detection(df_model, outliers_fraction, anomaly_columns):
        """ Brain of this anomaly detection
        Parameters
        ----------
        df: pd.DataFrame
            Dataframe containing x-axis (age_at_collection, which we call y), y-axis (MMI, which we call y_pred), and unique sample IDs (sampleID)
        """
        # Subset the dataframe by desired columns
        dataframe_filtered_columns = df_model[anomaly_columns]
        
        # Scale the column that we want to flag for anomalies
        min_max_scaler = StandardScaler()
        np_scaled = min_max_scaler.fit_transform(dataframe_filtered_columns)
        scaled_time_series = pd.DataFrame(np_scaled)
        
        # train isolation forest 
        model =  IsolationForest(contamination=outliers_fraction)
        model.fit(scaled_time_series)
        
        #Generate column for Isolation Forest-detected anomalies
        df_model["if_anomaly"] = model.predict(scaled_time_series)
        df_model["if_anomaly"] = df_model["if_anomaly"].map( {1: False, -1: True} )
        return df_model[df_model["if_anomaly"]==True].sampleID.values.tolist()
    
    X, y = df2vectors(df, feature_columns)
    y_pred = estimator.predict(X)
    df_model = pd.DataFrame(data={"y":y, "y_pred":y_pred, 'sampleID': df.sampleID})
    df_model = df_model.sort_values(by="y")
    df_model["y_pred_rolling_avg"] = df_model["y_pred"].rolling(window=window, min_periods=1, center=True).mean()
    df_model["y_pred_zscore"]= df_model["y_pred"]-df_model["y_pred_rolling_avg"]
    
    outliers = _inner_if_anomaly_detection(df_model, outliers_fraction, anomaly_columns=anomaly_columns)
        
    return outliers    



def svm_anomaly_detection(estimator, df, feature_columns, outliers_fraction, window, anomaly_columns):
    """
    Isolation Forest (IF) to detect anomalies in a time series. We unite all longitudinal data into one signal. 
    If we had more smaples per subjects, we could do this lpf anomaly detection for every subject separately.
    Parameters
    ----------
    df: pd.DataFrame
        Dataframe containing data
    num_of_stds: float
        Number of standard deviations of z-score that will be considered anomalous if higher.
    window: int
        Number of observations used for calculating the statistic.
    anomaly_columns: list
        Two options: `y_pred`, which ignores the timebox, and `y_pred_zcore`, or both.
    
    Returns
    -------
    outliers :list
        List of sampleIDs that are anomalous.
    """
    def _inner_svm_anomaly_detection(df, outliers_fraction, anomaly_columns):
        """ Brain of this anomaly detection
        Parameters
        ----------
        df: pd.DataFrame
            Dataframe containing x-axis (age_at_collection, which we call y), y-axis (MMI, which we call y_pred), and unique sample IDs (sampleID)
        """
        # Subset the dataframe by desired columns
        dataframe_filtered_columns = df[anomaly_columns]
        
        # Scale the column that we want to flag for anomalies
        min_max_scaler = StandardScaler()
        np_scaled = min_max_scaler.fit_transform(dataframe_filtered_columns)
        scaled_dataframe = pd.DataFrame(np_scaled)
        
        # Remove any NaN from the dataframe
        scaled_dataframe =scaled_dataframe.dropna()
        
        # Train the One Class SVM
        model = svm.OneClassSVM(nu=outliers_fraction, kernel="rbf", gamma=0.01)
        model.fit(scaled_dataframe)
        
        # Generate column for detected anomalies
        scaled_dataframe["svm_anomaly"] = pd.Series(model.predict(scaled_dataframe)).map({1: False, -1: True})
        df["svm_anomaly"] = scaled_dataframe["svm_anomaly"]
        return df[df["svm_anomaly"]==True].sampleID.values.tolist()
    
    X, y = df2vectors(df, feature_columns)
    y_pred = estimator.predict(X)
    
    df_model = pd.DataFrame(data={"y":y, "y_pred":y_pred, 'sampleID': df.sampleID})
    df_model = df_model.sort_values(by="y")
    df_model["y_pred_rolling_avg"] = df_model["y_pred"].rolling(window=window, min_periods=1, center=True).mean()
    df_model["y_pred_zscore"]= df_model["y_pred"]-df_model["y_pred_rolling_avg"]
    
    outliers = _inner_svm_anomaly_detection(df_model, outliers_fraction, anomaly_columns=anomaly_columns)
        
    return outliers    


import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt 
import shap
import numpy as np 
import pandas as pd 
import itertools
import seaborn as sns
import scipy.stats as stats
import math 
import pathlib
from matplotlib.patches import Patch
import scipy as sp
from microbiome.helpers import df2vectors
from microbiome.trajectory import plot_1_trajectory
from matplotlib.ticker import MaxNLocator
import plotly.graph_objects as go
from plotly.tools import mpl_to_plotly

def get_top_bacteria_in_time(estimator, df_all, top_bacteria, days_start, days_number_1unit, num_top_bacteria, average=np.median, std=np.std, time_unit_size=1, time_unit_name="days"):  # scipy.stats.hmean
    """
    e.g.
    time_type = "year", days_number_1unit = 12*30
    
    Parameters
    ----------
    num_top_bacteria: int
        Number of top important bacteria to be ploted in this time block.
    """
    df = df_all.copy()
    
    if "y" not in df.columns:
        X, y = df2vectors(df, top_bacteria)
        y_pred = estimator.predict(X)
        df["y"] = y
        df["y_pred"] = y_pred

    # from main df, extract only the time block we are interested in
    df_time_block = df[(days_start<df.y)&(df.y<days_start+days_number_1unit)]
    # and now exctract only the columns we need
    df_time_block = df_time_block[top_bacteria+["age_at_collection", "y", "y_pred"]]
    
    xpos, bottom, bacteria_name, ratios, avgs, stds = None, None, None, None, None, None
    feature_importance = None
    if df_time_block.shape[0]!=0:
        # average, std, and median of samples in the time block for a given bacteria (and age, y, y_pred)
        avgs    = pd.Series(data=dict([(c, average(df_time_block[c].values)) for c in df_time_block.columns])) 
        stds    = pd.Series(data=dict([(c, std(df_time_block[c].values)) for c in df_time_block.columns]))
         
        # for this time block average, determine what are the most important features based on SHAP
        X, y = df2vectors(pd.DataFrame(dict(avgs), index=[0]), top_bacteria)
        explainer = shap.TreeExplainer(estimator)
        shap_values = explainer.shap_values(X)
        feature_importance = pd.DataFrame(list(zip(top_bacteria, np.abs(shap_values).mean(0))), columns=['bacteria_name','feature_importance_vals'])
        #feature_importance.sort_values(by=['feature_importance_vals'], ascending=True,inplace=True)
        # besides bacteria name and its importance, add its average and std for this time block
        feature_importance["bacteria_avg"] = feature_importance.apply(lambda x: avgs[x["bacteria_name"]], axis=1)
        feature_importance["bacteria_std"] = feature_importance.apply(lambda x: stds[x["bacteria_name"]], axis=1)
        feature_importance.sort_values(by=['feature_importance_vals'], ascending=True,inplace=True)

        # location of boxplots (medians)
        xpos   = (days_start+days_number_1unit/2)/time_unit_size
        bottom = df_time_block.median()["y_pred"]/time_unit_size
        ratios = feature_importance["feature_importance_vals"].values[-num_top_bacteria:]
        bacteria_name = feature_importance["bacteria_name"].values[-num_top_bacteria:]
        ratios /= sum(ratios)
        
    return xpos, bottom, bacteria_name, ratios, avgs, stds, feature_importance, len(df_time_block)

def plot_importance_boxplots_over_age(estimator, df_all, top_bacteria, nice_name, units, num_top_bacteria=5, start_age=0, limit_age=None, patent=False, highlight_outliers=None, df_new=None, 
                                        time_unit_size=1, time_unit_name="days", box_height=None, img_file_name=None, plateau_area_start=None, longitudinal_mode="markers+lines", longitudinal_showlegend=True, 
                                        fillcolor_alpha=0.3, layout_settings=None, website=False, PI_percentage=90, dtick=2):
    
    df = df_all.copy()
    
    # if there are samples after units, add the last box at the end that would fill in 
    limit_age_max = limit_age or max(df.age_at_collection.astype(int))+1
    if sum(units) < limit_age_max:
        units.append(limit_age_max-sum(units))
    if not plateau_area_start:
        plateau_area_start = limit_age_max
    
    X, y = df2vectors(df, top_bacteria)
    y_pred = estimator.predict(X)

    df["y"] = y
    df["y_pred"] = y_pred


    if df_new is not None:
        X, y = df2vectors(df_new, top_bacteria)
        y_pred = estimator.predict(X)

        df_new["y"] = y
        df_new["y_pred"] = y_pred

    layout_settings_default = dict(
        height=900, 
        width=1100,
        barmode='stack', 
        uniformtext=dict(mode="hide", minsize=10),
        plot_bgcolor='rgba(0,0,0,0)', 
        paper_bgcolor='rgba(0,0,0,0)', 
        margin=dict(l=0, r=0, b=0, pad=0),
        title_text="Trajectory time blocks with top important bacteria"
    )

    if patent:
        palette = itertools.cycle(sns.color_palette("Greys", n_colors=11).as_hex())
        boxplot_alpha = 1.0
        outlier_color = "0,0,0"
        traj_color = "0,0,0"
        marker_outlier=dict(size=25, color=f'rgba({outlier_color},0.95)', symbol="star-open", line_width=4)
        layout_settings_default["font"] = dict(
                #family="Courier New, monospace",
                size=20,
                #color="RebeccaPurple"
            )
    else:
        palette = itertools.cycle(sns.color_palette("Paired").as_hex())
        boxplot_alpha = 0.8
        outlier_color = "255,0,0"
        traj_color = "26,150,65" 
        marker_outlier=dict(size=25, color=f'rgba({outlier_color},0.95)', symbol="star-open", line_width=4)    

    if layout_settings is None:
        layout_settings = {}
    layout_settings_final = {**layout_settings_default, **layout_settings}
        
    colors = {}
    ypos_text_margin = 1.5
    text_fontsize_samples = 15
    str_fmt = ".5f"
    latest_day = start_age
    legend_vals = []

    fig = go.Figure()  
    fig, ret_val, mae, r2, pi_median, traj_x, traj_pi, traj_mean = plot_1_trajectory(fig, estimator, df, top_bacteria, limit_age_max, time_unit_size, time_unit_name, traj_color=traj_color, traj_label="reference", 
                                                                   plateau_area_start=plateau_area_start, limit_age_max=limit_age_max, longitudinal_mode=longitudinal_mode, 
                                                                   longitudinal_showlegend=longitudinal_showlegend, fillcolor_alpha=fillcolor_alpha, highlight_outliers=highlight_outliers,
                                                                  marker_outlier=marker_outlier, df_new=df_new, plot_CI=False, plot_PI=True, PI_percentage=PI_percentage)

    if box_height is None:
        box_height = 2*np.median(traj_pi)
        print("box_height", box_height)

    for days_number_1unit in units:

        box_width=0.9 * days_number_1unit/time_unit_size

        xpos, bottom, bacteria_name, ratios, bacteria_avgs, bacteria_stds, _, samples_num = get_top_bacteria_in_time(estimator, df, top_bacteria, days_start=latest_day, days_number_1unit=days_number_1unit, num_top_bacteria=num_top_bacteria, time_unit_size=time_unit_size, time_unit_name=time_unit_name)
        if xpos is not None:
            bottom -= box_height/2
            xpos   -= box_width/2

            for j in range(len(ratios)):
                # text mean and std for this bacteria inside one boxplot
                bacteria_avg = bacteria_avgs[bacteria_name[j]]
                bacteria_std  = bacteria_stds[bacteria_name[j]]

                if not colors.get(bacteria_name[j], None):
                    colors[bacteria_name[j]] = next(palette)

                height = ratios[j]*box_height
                legendgroup = bacteria_name[j]
                if legendgroup not in legend_vals:
                    legend_vals.append(legendgroup)
                    showlegend = True
                else:
                    showlegend = False
                fig.add_trace(go.Bar(x=[xpos], y=[height],   #red
                        base=bottom,
                        offset=0,
                        width=box_width,
                        marker_color=colors[bacteria_name[j]],  # color dep on bacteria
                        legendgroup=legendgroup,
                        showlegend=showlegend,
                        name=nice_name(bacteria_name[j]),
                        hovertemplate="<br>".join([
                        f"<b>bacteria: {nice_name(bacteria_name[j])}</b>",
                        f"avg ± std: {bacteria_avg:{str_fmt}} ± {bacteria_std:{str_fmt}}",
                        f"importance: {ratios[j]*100:.2f}%",
                        f"# samples: {samples_num}",
                    ])))

                bottom += height

        latest_day += days_number_1unit

    
    fig.update_xaxes( 
        title=f"Age [{time_unit_name}]",
        range=(0, latest_day/time_unit_size),
        tick0=start_age/time_unit_size, dtick=dtick,
        showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey')
    fig.update_yaxes(
        title=f"Microbiome Maturation Index [{time_unit_name}]",
        tick0=start_age/time_unit_size, dtick=dtick,
        range=(0, latest_day/time_unit_size), 
        showline=True, linecolor='lightgrey', gridcolor='lightgrey') 

    fig.update_layout(**layout_settings_final)

    if img_file_name:
        pathlib.Path('/'.join(img_file_name.split('/')[:-1])).mkdir(parents=True, exist_ok=True)
        if "html" in img_file_name.lower(): 
            fig.write_html(img_file_name)
        elif any(extension in img_file_name for extension in ["png", "jpeg", "webp", "svg", "pdf", "eps" ]):
            fig.write_image(img_file_name)
        else:
            raise Exception(f"Extension {img_file_name.split('.')[-1]} is not implemented")
    

    if not website:
        fig.show()
    
    return fig, traj_x, traj_pi, traj_mean


def outlier_intervention(outlier_sampleID, estimator, df_all, feature_columns, nice_name, max_num_bacterias_to_change, traj_x, traj_mean, traj_pi, time_unit_size=30, time_unit_name="months", plot=False, file_name=None, output_html=False, normal_vals=[False],average=np.median, std=np.std):
    
    def prediction_interval_with_direction(x, traj_mean, traj_pi, direction):
        # direction=-1 for lower limit, +1 for upper
        p, _ = np.polyfit(traj_x, traj_mean+direction*traj_pi, 2, cov=True)    
        y = np.polyval(p, x) 
        return y
    
    ### Outlier importances before ###
    df_all = df_all.copy()
    # chosing one outlier
    df_all = df_all.reset_index(drop=True)
    i = df_all.loc[df_all.sampleID==outlier_sampleID].index[0]
    
    X, y = df2vectors(df_all, feature_columns)
    y_pred = estimator.predict(X)

    explainer = shap.TreeExplainer(estimator)
    shap_values = explainer.shap_values(X)

    # margin will be used as a +- range we want to put the outlier in
    # for e.g. outlier has 15 months on x-axis, so in order to align it to healthy region on y-axis, we need to collect the samples in the region 15+-margin months and do the stats on these healthy subset of samples.
    # if outlier has age_at_collection < 7 months, put +-1 month in MMI, etc.
    margin = 1*30/time_unit_size if int(round(y[i]/time_unit_size)) < 7*30/time_unit_size else 2*30/time_unit_size if  7*30/time_unit_size <= int(round(y[i]/time_unit_size)) < 27*30/time_unit_size else int(round(np.median(traj_pi)))
    #print(margin)

    shift_direction = -1 if y_pred[i]>y[i] else +1
    additional_push = int(shift_direction*np.median(traj_pi)/2)
    left  = max(int(round(y[i]/time_unit_size - margin)) + additional_push, 0)
    right = max(int(round(y[i]/time_unit_size + margin)) + additional_push,0) + margin
    print(left, right)

    left_pred  = max(int(round(prediction_interval_with_direction(right, traj_mean, traj_pi, direction=-1))), 0)
    right_pred     = int(round(prediction_interval_with_direction(left, traj_mean, traj_pi, direction=+1)))
    #print(left_pred, right_pred)

    ret_val1 = "\n--- Before Intervention ---\n"
    ret_val1 += f"age_at_collection = {y[i]:.2f} days [{int(round(y[i]/time_unit_size))} {time_unit_name}]\n"
    ret_val1 += f"microbiota_age = {y_pred[i]:.2f} days [{int(round(y_pred[i]/time_unit_size))} {time_unit_name}]\n"
    ret_val1 += f"put in prediction interval (y-axis) between {left_pred} [{left_pred*time_unit_size} days] and {right_pred} [{right_pred*time_unit_size} days] {time_unit_name} to make it healthy\n"
    ret_val1 += f"do the stats on interval (x-axis) between {left} [{left*time_unit_size} days] and {right} [{right*time_unit_size} days] {time_unit_name}\n"
    if not output_html:
        print(ret_val1)

    # shorten bacteria names just for the force plot below
    feature_columns_short = list(map(nice_name, feature_columns))
    if plot:
        shap.force_plot(explainer.expected_value, shap_values[i], features=X[i], feature_names=feature_columns_short, text_rotation=90, matplotlib=True)
        plt.show()
    else:
        fig1, ax = plt.subplots()
        shap.force_plot(explainer.expected_value, shap_values[i], features=X[i], feature_names=feature_columns_short, text_rotation=90, matplotlib=plot)
        
        fig1 =  mpl_to_plotly(fig1)
            
        fig1.update_xaxes(title="SHAP value (impact on model output)", 
                        showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray') 
        fig1.update_yaxes(title="Features", 
                            tickmode='array',
                            tickvals=list(range(0, min(20, len(feature_columns_short)))),  #list(X_train.columns[np.argsort(np.abs(shap_values[0]).mean(0))])[::-1][:min(20, len(feature_names))], # 
                            #ticktext=list(map(nice_name, list(X_train.columns[np.argsort(np.abs(shap_values[0]).mean(0))])[::-1][:min(20, len(feature_names))][::-1])),
                        showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray')  
        fig1.update_layout(#height=layout_height, width=layout_width,
                        #paper_bgcolor="white",#'rgba(0,0,0,0)', 
                        plot_bgcolor='rgba(0,0,0,0)', 
                        margin=dict(l=0, r=0, b=0, pad=0),
                        title_text="Classification Important Features")

    ### Intervention Table ###
    # get importances of outlier
    feature_importance_outlier = pd.DataFrame(list(zip(feature_columns, np.abs(shap_values).mean(0))), columns=['bacteria_name','feature_importance_outlier'])
    feature_importance_outlier.sort_values(by=['feature_importance_outlier'], ascending=True, inplace=True)

    # get importances of the trajectory
    _, _, _, _, _, _, feature_importance, _ = get_top_bacteria_in_time(estimator, df_all, feature_columns, 
                                                                       days_start=left*time_unit_size, 
                                                                       days_number_1unit=(left+right)*time_unit_size, 
                                                                       num_top_bacteria=5, 
                                                                       average=average, std=std)
    print(feature_importance)
    #display(feature_importance)
    # merge two importances
    feature_importance["outlier"] = feature_importance.apply(lambda x: df_all.iloc[i][x["bacteria_name"]], axis=1)
    feature_importance["normal"] = feature_importance.apply(lambda x: True if x["bacteria_avg"]-x["bacteria_std"] < x["outlier"] < x["bacteria_avg"]+x["bacteria_std"] else False, axis=1)
    feature_importance = pd.merge(feature_importance, feature_importance_outlier, on="bacteria_name")
    feature_importance.sort_values(by=['feature_importance_outlier', 'feature_importance_vals'], ascending=False,inplace=True)
    if file_name:
        feature_importance.to_csv(file_name, index=False)
    
    
    # change the most important bacteria value
    # Faecalibacterium is the first important bacteria that is outside the healthy range
    df_all_updated = df_all.copy()

    #bacterias_to_change = feature_importance.iloc[:5].bacteria_name.values  #
    #bacterias_to_change = list(set(list(feature_importance[feature_importance.normal==False].bacteria_name.values) + list( feature_importance.iloc[:5].bacteria_name.values)))
    bacterias_to_change = feature_importance[feature_importance.normal.isin(normal_vals)].bacteria_name.values[:max_num_bacterias_to_change]

    ret_val2 = "\n--- Intervention bacteria ---\n"
    for bacteria_to_change in bacterias_to_change:
        ret_val2 += f"\n{nice_name(bacteria_to_change)}\n"
        ret_val2 +=f"before = {df_all_updated.at[i, bacteria_to_change]}\n"
        df_all_updated.at[i, bacteria_to_change] = feature_importance[feature_importance.bacteria_name==bacteria_to_change]["bacteria_avg"].values[0]
        ret_val2 +=f"after = {df_all_updated.at[i, bacteria_to_change]}\n"
    if not output_html:
        print(ret_val2)

    X, y = df2vectors(df_all_updated, feature_columns)
    y_pred = estimator.predict(X)

    explainer = shap.TreeExplainer(estimator)
    shap_values = explainer.shap_values(X)

    margin = 1*30/time_unit_size if int(round(y[i]/time_unit_size)) < 7*30/time_unit_size else 2*30/time_unit_size if  7*30/time_unit_size <= int(round(y[i]/time_unit_size)) < 27*30/time_unit_size else int(round(np.median(traj_pi)))
    
    shift_direction = -1 if y_pred[i]>y[i] else +1
    additional_push = shift_direction*np.median(traj_pi)/2
    left  = max(int(round(y[i]/time_unit_size - margin)) + additional_push, 0)
    right = max(int(round(y[i]/time_unit_size + margin)) + additional_push,0) + margin
    print(left, right)
    
    left_pred  = max(int(round(prediction_interval_with_direction(right, traj_mean, traj_pi, direction=-1))), 0)
    right_pred     = int(round(prediction_interval_with_direction(left, traj_mean, traj_pi, direction=+1)))
    
    ret_val3 = "\n--- After Intervention ---\n"
    ret_val3 += f"age_at_collection = {y[i]:.2f} days [{int(round(y[i]/time_unit_size))} {time_unit_name}]\n"
    ret_val3 += f"microbiota_age = {y_pred[i]:.2f} days [{int(round(y_pred[i]/time_unit_size))} {time_unit_name}]\n"
    ret_val3 += f"put in prediction interval (y-axis) between {left_pred} [{left_pred*time_unit_size} days] and {right_pred} [{right_pred*time_unit_size} days] months to make it healthy\n"
    ret_val3 += f"do the stats on interval (x-axis) between {left} [{left*time_unit_size} days] and {right} [{right*time_unit_size} days] {time_unit_name}\n"
    if not output_html:
        print(ret_val3)
    if plot:
        shap.force_plot(explainer.expected_value, shap_values[i], features=X[i], feature_names=feature_columns_short, text_rotation=90, matplotlib=True)
        plt.show()
    else:

        fig2, ax = plt.subplots()
        shap.force_plot(explainer.expected_value, shap_values[i], features=X[i], feature_names=feature_columns_short, text_rotation=90, matplotlib=plot)
        
        fig2 =  mpl_to_plotly(fig2)
            
        fig2.update_xaxes(title="SHAP value (impact on model output)", 
                        showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray') 
        fig2.update_yaxes(title="Features", 
                            tickmode='array',
                            tickvals=list(range(0, min(20, len(feature_columns_short)))),  #list(X_train.columns[np.argsort(np.abs(shap_values[0]).mean(0))])[::-1][:min(20, len(feature_names))], # 
                            #ticktext=list(map(nice_name, list(X_train.columns[np.argsort(np.abs(shap_values[0]).mean(0))])[::-1][:min(20, len(feature_names))][::-1])),
                        showline=True, linecolor='lightgrey', gridcolor='lightgrey', zeroline=True, zerolinecolor='lightgrey', showspikes=True, spikecolor='gray')  
        fig2.update_layout(#height=layout_height, width=layout_width,
                        #paper_bgcolor="white",#'rgba(0,0,0,0)', 
                        plot_bgcolor='rgba(0,0,0,0)', 
                        margin=dict(l=0, r=0, b=0, pad=0),
                        title_text="Classification Important Features")
    
    
    return df_all_updated, fig1, fig2, ret_val1+ret_val2+ret_val3